def register_metric_sections(card, registry):
    """
    Registers metric sections for a model, specifically the training metric and sample weights.
    If sample weights are not used, then that particular section is skipped.
    """

    @registry.register("primary_metric_section")
    def primary_metric():
        """
        Returns a markdown string describing the primary metric used for training the model. This section
        is meant to be more verbose in the model card, explaining what the metric is for a non-technical audience.
        """
        metric_map = {
            "log_loss": f"{card.format.indent_level(1)}- Our primary metric for training was log loss to ensure that the model produces well-calibrated probability estimates.\n{card.format.indent_level(1)}- Lower log loss is better, as it indicates more accurate and confident probability predictions.",
            "recall": f"{card.format.indent_level(1)}- Our primary metric for training was recall in order to ensure that we correctly identify as many students in need of support as possible.\n{card.format.indent_level(1)}- Higher recall is better, as it indicates fewer students in need are missed.",
            "precision": f"{card.format.indent_level(1)}- Our primary metric for training was precision to ensure that when the model identifies a student as needing support, it is likely to be correct.\n{card.format.indent_level(1)}- Higher precision is better, as it indicates fewer students are incorrectly flagged.",
            "roc_auc": f"{card.format.indent_level(1)}- Our primary metric for training was ROC AUC, which measures the model's ability to distinguish between students who need support and those who do not.\n{card.format.indent_level(1)}- Higher ROC AUC is better, as it indicates stronger overall classification performance across all thresholds.",
            "f1": f"{card.format.indent_level(1)}- Our primary metric for training was F1-score to balance the trade-off between precision and recall.\n{card.format.indent_level(1)}- A higher F1-score indicates that the model is effectively identifying students in need while minimizing both false positives and false negatives.",
        }
        metric = card.cfg.modeling.training.primary_metric
        return metric_map.get(
            metric, f"{card.format.indent_level(1)}- Default metric explanation."
        )

    @registry.register("sample_weight_section")
    def sample_weight():
        """
        Returns a markdown string describing the sample weights used for training the model. This section
        will still print out how many experiments were run but the sample weight details are optional, depending
        on where a column with a substring of "sample_weight" exists in the training data.
        """
        used_weights = any(
            col.startswith("sample_weight") for col in card.training_data.columns
        )
        sw_note = (
            f"{card.format.indent_level(1)}- Sample weights were used to stabilize training."
            if used_weights
            else None
        )
        num_runs = card.context.get("num_runs_in_experiment", None)
        if isinstance(num_runs, int) and num_runs > 0:
            mlops_note = (
                f"{card.format.indent_level(1)}- Utilizing Databricks and AutoML, "
                f"we initiated an MLOps pipeline for data processing and model experimentation, "
                f"processing {num_runs} different machine-learning models to optimize our model."
            )
        else:
            mlops_note = (
                f"{card.format.indent_level(1)}- Utilizing Databricks and AutoML, "
                f"we initiated an MLOps pipeline for data processing and model experimentation."
            )
        return "\n".join(filter(None, [mlops_note, sw_note]))



import pytest
from student_success_tool.reporting.utils.formatting import Formatting


@pytest.fixture
def formatter():
    return Formatting()


@pytest.mark.parametrize(
    "input_text,expected",
    [
        ("bachelor's degree", "Bachelor's Degree"),
        ("full-time", "Full-Time"),
        ("associate's-degree", "Associate's-Degree"),
        ("first_generation", "First Generation"),
        ("1st_semester", "1st Semester"),
        ("STEM", "Stem"),  # no acronym preservation for now
        (0.0, "0.0"),
        (3, "3"),
        ("3.0", "3.0"),
        ("4", "4"),
    ],
)
def test_friendly_case_capitalize_true(formatter, input_text, expected):
    result = formatter.friendly_case(input_text)
    assert result == expected


@pytest.mark.parametrize(
    "input_text,expected",
    [
        ("bachelor's degree", "bachelor's degree"),
        ("full-time", "full-time"),
        ("first_term", "first term"),
        ("1st_semester", "1st semester"),
    ],
)
def test_friendly_case_capitalize_false(formatter, input_text, expected):
    result = formatter.friendly_case(input_text, capitalize=False)
    assert result == expected



# Databricks notebook source
# %pip install faker pandas


# COMMAND ----------

num_records = 100  # how many records to generate

columns = [
    "StuNum",
    "First_Enrollment_Date",
    "FirstGenFlag",
    "AgeGroup",
    "Gender",
    "Race",
    "Ethnicity",
    "IncarceratedFlag",
    "DualEnrollFlag",
    "ABEFlag",
    "Zip",
    "DisabilityFlag",
    "HighSchoolStatus",
    "HighSchoolGradDate",
    "HighSchoolGPA",
    "StudentAthlete",
    "EarnedAward",
    "AwardsEarned",
]

# Define choices for each column
stu_num_range = range(10000, 99999)  # Unique student numbers
first_gen_choices = ["Y", "N", "U"]
age_group_choices = [
    "Under 18",
    "18-19",
    "20-24",
    "25-29",
    "30-39",
    "40-49",
    "50-64",
    "65+",
]
gender_choices = ["Female", "Male", "Not Specified"]
race_choices = ["Non-Hispanic", "Hispanic"]
ethnicity_choices = [
    "Two or more races",
    "Caucasian/White",
    "Unknown",
    "Black or African American",
    "American Indian or Alaska Native",
    "Asian",
    "Native Hawaiian or Other Pacific Islander",
    "Nonresident Alien",
]
incarcerated_choices = [0, 1]
dual_enroll_choices = [0, 1]
abe_choices = [0, 1]
disability_choices = ["N", "Y"]
high_school_status_choices = [
    "TraditionalHighSchoolDiploma",
    "No High School Record",
    "GED",
]


# COMMAND ----------

import random
from datetime import datetime

import pandas as pd
from faker import Faker

# Initialize Faker
fake = Faker()


# Function to generate random dates
def random_date(start, end):
    return fake.date_between(start_date=start, end_date=end).strftime("%m/%d/%Y")


# Define the parameters for the dataset
data = []

for _ in range(num_records):
    stu_num = fake.unique.random_int(min=10000, max=99999)  # Unique student number
    first_enrollment_date = random_date(datetime(2000, 1, 1), datetime.now())
    first_gen_flag = random.choice(first_gen_choices)
    age_group = random.choice(age_group_choices)
    gender = random.choice(gender_choices)
    race = random.choice(race_choices)
    ethnicity = random.choice(ethnicity_choices)
    incarcerated_flag = random.choice(incarcerated_choices)
    dual_enroll_flag = random.choice(dual_enroll_choices)
    abe_flag = random.choice(abe_choices)
    zip_code = fake.zipcode()  # Generate arbitrary ZIP code
    disability_flag = random.choice(disability_choices)
    high_school_status = random.choice(high_school_status_choices)
    high_school_grad_date = random.choice(
        [random_date(datetime(1990, 1, 1), datetime.now()), None]
    )
    high_school_gpa = random.choice(
        [None] * 80 + [round(random.uniform(0, 4), 2) for _ in range(20)]
    )  # 80% nulls
    student_athlete = random.choice(["Y", "N"])
    earned_award = random.choice([None] * 80 + ["Y" for _ in range(20)])  # 80% nulls
    awards_earned = random.choice(
        [None] * 80 + [random.randint(1, 10) for _ in range(20)]
    )  # 80% nulls

    data.append(
        [
            stu_num,
            first_enrollment_date,
            first_gen_flag,
            age_group,
            gender,
            race,
            ethnicity,
            incarcerated_flag,
            dual_enroll_flag,
            abe_flag,
            zip_code,
            disability_flag,
            high_school_status,
            high_school_grad_date,
            high_school_gpa,
            student_athlete,
            earned_award,
            awards_earned,
        ]
    )
df = pd.DataFrame(data, columns=columns)

# Save to CSV
df.to_csv("DataKind_School1_StudentFile.csv", index=False)


# COMMAND ----------

fake = Faker()

# Define choices for each column
semesters = [
    "2018FA",
    "2018SP",
    "2019FA",
    "2019SP",
    "2020FA",
    "2020SP",
    "2021FA",
    "2021SP",
]
enrollment_types = ["PT", "FT"]
degree_types = [
    "Associate of Science",
    "Associate of Applied Science",
    "Associate of Arts",
    "Associate of Business",
    "Associate of General Studies",
    "Certificate",
    "Non-Academic",
]
pell_recipient_choices = [None, "Y"]
majors = [
    "Custodian Studies Certificate",
    "Construction Technician",
    "Agriculture Associate of Arts",
    "Business Administration",
    "Nursing",
    "Computer Science",
    "Graphic Design",
]
department = [None]  # All NULL

columns = [
    "Student_ID",
    "Semester",
    "Enrollment_Type",
    "Degree_Type_Pursued",
    "Intent_To_Transfer_Flag",
    "Pell_Recipient",
    "Major",
    "Department",
    "Cumulative_GPA",
    "Semester_GPA",
    "Number_Of_Courses_Enrolled",
    "Number_Of_Credits_Attempted",
    "Number_Of_Credits_Failed",
    "Number_Of_Credits_Earned",
    "Course_Pass_Rate",
    "Online_Course_Rate",
]


# Function to enforce GPA and credit rules
def generate_valid_enrollment_data():
    number_of_courses = random.randint(0, 10)
    if number_of_courses == 0:
        number_of_credits_attempted = 0
        number_of_credits_failed = 0
        number_of_credits_earned = 0
    else:
        number_of_credits_attempted = random.randint(
            number_of_courses, 30
        )  # Ensure at least as many credits as courses
        number_of_credits_failed = random.randint(0, number_of_credits_attempted)
        number_of_credits_earned = (
            number_of_credits_attempted - number_of_credits_failed
        )

    cumulative_gpa = round(random.uniform(0, 4), 2)
    semester_gpa = round(random.uniform(0, 4), 2)
    course_pass_rate = round(random.uniform(0, 1), 2)
    online_course_rate = round(random.uniform(0, 1), 2)

    return [
        number_of_courses,
        number_of_credits_attempted,
        number_of_credits_failed,
        number_of_credits_earned,
        cumulative_gpa,
        semester_gpa,
        course_pass_rate,
        online_course_rate,
    ]


data = []

for _ in range(num_records):
    student_id = fake.unique.random_int(min=10000, max=99999)  # Unique student ID
    semester = random.choice(semesters)
    enrollment_type = random.choice(enrollment_types)
    degree_type = random.choice(degree_types)
    intent_to_transfer_flag = random.choice(
        ["Y", "N"]
    )  # Not specified in requirements but added for completeness
    pell_recipient = random.choice(pell_recipient_choices)
    major = random.choice(majors)
    dept = random.choice(department)  # All NULL

    # Generate valid enrollment data
    (
        number_of_courses,
        number_of_credits_attempted,
        number_of_credits_failed,
        number_of_credits_earned,
        cumulative_gpa,
        semester_gpa,
        course_pass_rate,
        online_course_rate,
    ) = generate_valid_enrollment_data()

    data.append(
        [
            student_id,
            semester,
            enrollment_type,
            degree_type,
            intent_to_transfer_flag,
            pell_recipient,
            major,
            dept,
            cumulative_gpa,
            semester_gpa,
            number_of_courses,
            number_of_credits_attempted,
            number_of_credits_failed,
            number_of_credits_earned,
            course_pass_rate,
            online_course_rate,
        ]
    )

df = pd.DataFrame(data, columns=columns)

# Save to CSV
df.to_csv("DataKind_School1_SemesterFile.csv", index=False)


# COMMAND ----------

fake = Faker()

columns = [
    "Student_ID",
    "Semester",
    "Course_Prefix",
    "Course_Number",
    "Course_Type",
    "Pass/Fail_Flag",
    "Grade",
    "Prerequisite_Course_Flag",
    "Online_Course_Flag",
    "Number_Of_Credits_Attempted",
    "Number_Of_Credits_Earned",
    "Modality",
    "Weeks",
]

# Define choices for each column
course_prefixes = ["BIO", "COM", "ENG", "MAT", "HIS", "CHE", "PHY", "ART", "MUS", "PSY"]
course_types = ["Undergraduate"]  # All courses are undergraduate
pass_fail_choices = ["Y", "N"]
prerequisite_flag_choices = ["Y", None]
online_flag_choices = [1, 0]
modalities = [
    "Online",
    "Face to Face",
    "Online/Synch",
    "Hybrid - Online/iTV",
    "Hybrid - F2F/Online",
    "Independent Study",
    "Practicum",
]

# Load previously generated student IDs and semesters
semester_df = pd.read_csv("DataKind_School1_SemesterFile.csv")
student_ids = semester_df["Student_ID"].unique()
semesters = semester_df["Semester"].unique()

# Define the parameters for the dataset
data = []

# For each student and semester, generate multiple courses
for student_id in student_ids:
    # Get the semesters for this student
    student_semesters = semester_df[semester_df["Student_ID"] == student_id]

    for _, semester_row in student_semesters.iterrows():
        semester = semester_row["Semester"]
        number_of_courses = semester_row["Number_Of_Courses_Enrolled"]

        for _ in range(number_of_courses):
            course_prefix = random.choice(course_prefixes)
            course_number = random.randint(0, 299)
            course_type = random.choice(course_types)
            pass_fail_flag = random.choice(pass_fail_choices)
            grade = random.choice(
                ["A", "B", "C", "D", "F", None]
            )  # Include None for null grades
            prerequisite_flag = random.choice(prerequisite_flag_choices)
            online_course_flag = random.choice(online_flag_choices)
            number_of_credits_attempted = random.randint(
                1, 4
            )  # Assuming credits between 1 and 4
            number_of_credits_earned = random.choice(
                [number_of_credits_attempted, None]
            )  # Could be null
            modality = random.choice(modalities)
            weeks = random.randint(1, 16)  # Typical course duration in weeks

            data.append(
                [
                    student_id,
                    semester,
                    course_prefix,
                    course_number,
                    course_type,
                    pass_fail_flag,
                    grade,
                    prerequisite_flag,
                    online_course_flag,
                    number_of_credits_attempted,
                    number_of_credits_earned,
                    modality,
                    weeks,
                ]
            )

df = pd.DataFrame(data, columns=columns)

# Save to CSV
df.to_csv("DataKind_School1_CourseFile.csv", index=False)



import pandas as pd
import pytest

from student_success_tool.modeling import utils


@pytest.mark.parametrize(
    ["df", "label_fracs", "shuffle", "seed"],
    [
        (
            pd.DataFrame(data=list(range(1000))),
            {"train": 0.5, "test": 0.5},
            True,
            10,
        ),
        (
            pd.DataFrame(data=list(range(1000))),
            {"train": 0.6, "test": 0.2, "valid": 0.2},
            False,
            11,
        ),
        (
            pd.DataFrame(data=list(range(1000))),
            {"train": 0.5, "test": 0.5},
            True,
            42,
        ),
    ],
)
def test_compute_dataset_splits(df, label_fracs, shuffle, seed):
    obs = utils.compute_dataset_splits(
        df, label_fracs=label_fracs, shuffle=shuffle, seed=seed
    )
    assert isinstance(obs, pd.Series)
    assert len(obs) == len(df)
    labels = list(label_fracs.keys())
    fracs = list(label_fracs.values())
    obs_value_counts = obs.value_counts(normalize=True)
    exp_value_counts = pd.Series(
        data=fracs,
        index=pd.Index(labels, dtype="string", name="split"),
        name="proportion",
        dtype="Float64",
    )
    assert (
        pd.testing.assert_series_equal(
            obs_value_counts, exp_value_counts, rtol=0.15, check_like=True
        )
        is None
    )
    if seed is not None:
        obs2 = utils.compute_dataset_splits(
            df, label_fracs=label_fracs, shuffle=shuffle, seed=seed
        )
        assert obs.equals(obs2)


@pytest.mark.parametrize(
    ["df", "target_col", "class_weight", "exp"],
    [
        (
            pd.DataFrame({"target": [1, 1, 1, 0]}),
            "target",
            "balanced",
            pd.Series(
                [0.667, 0.667, 0.667, 2.0], dtype="float32", name="sample_weight"
            ),
        ),
        (
            pd.DataFrame({"target": [1, 1, 1, 0]}),
            "target",
            {1: 2, 0: 0.5},
            pd.Series([2.0, 2.0, 2.0, 0.5], dtype="float32", name="sample_weight"),
        ),
    ],
)
def test_compute_sample_weights(df, target_col, class_weight, exp):
    obs = utils.compute_sample_weights(
        df, target_col=target_col, class_weight=class_weight
    )
    assert isinstance(obs, pd.Series)
    assert len(obs) == len(df)
    assert pd.testing.assert_series_equal(obs, exp, rtol=0.01) is None



import logging
import time
import typing as t

import pandas as pd

# TODO: how do we specify databricks.automl as a package dependency?
# i.e. what's the package name??

LOGGER = logging.getLogger(__name__)


def run_automl_classification(
    df: pd.DataFrame,
    *,
    target_col: str,
    primary_metric: str,
    institution_id: str,
    job_run_id: str,
    student_id_col: str,
    **kwargs: object,
) -> object:
    """
    Wrap :func:`databricks.automl.classify()` to allow testing and ensure that
    our parameters are used properly.

    Args:
        df: Dataset containing features and target used to train classifier model,
            as well as ``student_id_col`` and any other columns
            specified in the optional ``**kwargs``
        target_col: Column name for the target to be predicted
        primary_metric: Metric used to evaluate and rank model performance;
            currently supported classification metrics include "f1", "log_loss",
            "precision", "accuracy", "roc_auc"
        institution_id: Unique identifier for the dataset's institution,
            used to name the experiment in Databricks
        job_run_id: job run ID of Databricks workflow for labeling experiment
        student_id_col: column name containing student IDs to exclude from training.
        **kwargs: keyword arguments to be passed to :func:`databricks.automl.classify()`
            If time_col is provided, AutoML tries to split the dataset into training, validation,
            and test sets chronologically, using the earliest points as training data and the latest
            points as a test set. AutoML accepts timestamps and integeters. With Databricks Runtime
            10.2 ML and above, string columns are also supported using semantic detection. However,
            we have not found AutoML to accurately support string types relevant to our data,
            so our wrapper function requires that the column type is a timestamp or integer.

    Returns:
        AutoMLSummary: an AutoML object that describes and can be used to pull
            the metrics, parameters, and other details for each of the trials.

    References:
        - https://docs.databricks.com/en/machine-learning/automl/automl-api-reference.html#classify
    """
    if kwargs.get("time_col") is not None:
        time_col = kwargs["time_col"]
        if not (
            pd.api.types.is_datetime64_any_dtype(df[time_col].dtype)
            or pd.api.types.is_integer_dtype(df[time_col].dtype)
        ):
            raise ValueError(
                f"The time column specified ({time_col}) for splitting into training, "
                "testing, and validation datasets is not a datetime or integer, "
                f"rather it's of type {df[time_col].dtype}. Please revise!"
            )

    # set some sensible default arguments
    kwargs.setdefault("pos_label", True)
    # TODO: tune this! https://app.asana.com/0/0/1206779161097924/f
    kwargs.setdefault("timeout_minutes", 5)
    exclude_cols = kwargs.pop("exclude_cols", [])
    assert isinstance(exclude_cols, list)  # type guard
    exclude_cols = exclude_cols.copy()
    if student_id_col is not None and student_id_col not in exclude_cols:
        exclude_cols.append(student_id_col)

    experiment_name = get_experiment_name(
        institution_id=institution_id,
        job_run_id=job_run_id,
        primary_metric=primary_metric,
        timeout_minutes=kwargs["timeout_minutes"],  # type: ignore
        exclude_frameworks=kwargs.get("exclude_frameworks"),  # type: ignore
    )

    from databricks import automl  # type: ignore  # importing here for mocking in tests

    LOGGER.info("running experiment: %s ...", experiment_name)
    summary = automl.classify(
        dataset=df,
        target_col=target_col,
        primary_metric=primary_metric,
        experiment_name=experiment_name,
        exclude_cols=exclude_cols,
        **kwargs,
    )
    return summary


def get_experiment_name(
    *,
    institution_id: str,
    job_run_id: str,
    primary_metric: str,
    timeout_minutes: int,
    exclude_frameworks: t.Optional[list[str]] = None,
) -> str:
    """
    Get a descriptive experiment name based on more important input parameters.

    See Also:
        - :func:`run_automl_classification()`

    References:
        - https://docs.databricks.com/en/machine-learning/automl/automl-api-reference.html#classify
    """
    name_components = [
        institution_id,
        f"{job_run_id=}",
        f"{primary_metric=}",
        f"{timeout_minutes=}",
    ]
    if exclude_frameworks:
        name_components.append(f"exclude_frameworks={','.join(exclude_frameworks)}")
    name_components.append(time.strftime("%Y-%m-%dT%H:%M:%S"))

    name = "__".join(name_components)
    if len(name) > 500:
        LOGGER.warning("truncating long experiment name '%s' to first 500 chars", name)
        name = name[:500]
    return name



import functools as ft

import pandas as pd
import pytest

from student_success_tool.preprocessing import checkpoints
from student_success_tool.preprocessing.targets.pdp import credits_earned


@pytest.mark.parametrize(
    [
        "df",
        "min_num_credits",
        "checkpoint",
        "intensity_time_limits",
        "num_terms_in_year",
        "max_term_rank",
        "exp",
    ],
    [
        # base case
        (
            pd.DataFrame(
                {
                    "student_id": ["01", "01", "01", "01"],
                    "enrollment_intensity": ["FT", "FT", "FT", "FT"],
                    "num_credits": [12, 24, 36, 48],
                    "term_rank": [1, 2, 3, 5],
                },
            ).astype({"student_id": "string", "enrollment_intensity": "string"}),
            45.0,
            # first term as checkpoint
            pd.DataFrame(
                {
                    "student_id": ["01"],
                    "enrollment_intensity": ["FT"],
                    "num_credits": [12],
                    "term_rank": [1],
                },
            ).astype({"student_id": "string", "enrollment_intensity": "string"}),
            {"FT": [4, "term"]},
            2,
            "infer",
            pd.Series(
                data=[False],
                index=pd.Index(["01"], dtype="string", name="student_id"),
                name="target",
                dtype="boolean",
            ),
        ),
        # multiple students, one full-time the other part-time
        (
            pd.DataFrame(
                {
                    "student_id": ["01", "01", "02", "02"],
                    "enrollment_intensity": ["FT", "FT", "PT", "PT"],
                    "num_credits": [12, 48, 8, 64],
                    "term_rank": [1, 4, 1, 10],
                },
            ).astype({"student_id": "string", "enrollment_intensity": "string"}),
            60.0,
            # first term as checkpoint
            pd.DataFrame(
                {
                    "student_id": ["01", "02"],
                    "enrollment_intensity": ["FT", "PT"],
                    "num_credits": [12, 8],
                    "term_rank": [1, 1],
                },
            ).astype({"student_id": "string", "enrollment_intensity": "string"}),
            {"FT": [4, "term"], "PT": [8, "term"]},
            2,
            "infer",
            pd.Series(
                data=[True, True],
                index=pd.Index(["01", "02"], dtype="string", name="student_id"),
                name="target",
                dtype="boolean",
            ),
        ),
        # checkpoint given as callable
        (
            pd.DataFrame(
                {
                    "student_id": ["01", "01", "02", "02"],
                    "enrollment_intensity": ["FT", "FT", "PT", "PT"],
                    "num_credits": [12, 48, 8, 64],
                    "term_rank": [1, 4, 1, 10],
                },
            ).astype({"student_id": "string", "enrollment_intensity": "string"}),
            60.0,
            # first term as checkpoint via callable
            ft.partial(
                checkpoints.pdp.first_student_terms,
                student_id_cols="student_id",
                sort_cols="term_rank",
                include_cols=["enrollment_intensity", "num_credits"],
            ),
            {"FT": [4, "term"], "PT": [8, "term"]},
            2,
            "infer",
            pd.Series(
                data=[True, True],
                index=pd.Index(["01", "02"], dtype="string", name="student_id"),
                name="target",
                dtype="boolean",
            ),
        ),
        # not enough terms in dataset to compute target
        (
            pd.DataFrame(
                {
                    "student_id": ["01", "01", "01", "01"],
                    "enrollment_intensity": ["FT", "PT", "PT", "FT"],
                    "num_credits": [12, 24, 36, 48],
                    "term_rank": [1, 2, 6, 7],
                },
            ).astype({"student_id": "string", "enrollment_intensity": "string"}),
            45.0,
            # second term as checkpoint
            pd.DataFrame(
                {
                    "student_id": ["01"],
                    "enrollment_intensity": ["PT"],
                    "num_credits": [24],
                    "term_rank": [3],
                },
            ).astype({"student_id": "string", "enrollment_intensity": "string"}),
            {"FT": [4, "term"], "PT": [8, "term"]},
            4,
            "infer",
            pd.Series(
                data=[],
                index=pd.Index([], dtype="string", name="student_id"),
                name="target",
                dtype="boolean",
            ),
        ),
        # one time limit for all enrollment intensities
        (
            pd.DataFrame(
                {
                    "student_id": ["01", "01", "02", "02"],
                    "enrollment_intensity": ["FT", "FT", "PT", "PT"],
                    "num_credits": [12, 48, 8, 64],
                    "term_rank": [1, 4, 1, 8],
                },
            ).astype({"student_id": "string", "enrollment_intensity": "string"}),
            60.0,
            # first term as checkpoint
            pd.DataFrame(
                {
                    "student_id": ["01", "02"],
                    "enrollment_intensity": ["FT", "PT"],
                    "num_credits": [12, 8],
                    "term_rank": [1, 1],
                },
            ).astype({"student_id": "string", "enrollment_intensity": "string"}),
            {"*": [8, "term"]},
            2,
            10,
            pd.Series(
                data=[True, False],
                index=pd.Index(["01", "02"], dtype="string", name="student_id"),
                name="target",
                dtype="boolean",
            ),
        ),
    ],
)
def test_compute_target(
    df,
    min_num_credits,
    checkpoint,
    intensity_time_limits,
    num_terms_in_year,
    max_term_rank,
    exp,
):
    obs = credits_earned.compute_target(
        df,
        min_num_credits=min_num_credits,
        checkpoint=checkpoint,
        intensity_time_limits=intensity_time_limits,
        num_terms_in_year=num_terms_in_year,
        max_term_rank=max_term_rank,
        student_id_cols="student_id",
        enrollment_intensity_col="enrollment_intensity",
        num_credits_col="num_credits",
        term_rank_col="term_rank",
    )
    assert isinstance(obs, pd.Series)
    assert pd.testing.assert_series_equal(obs, exp) is None



import typing as t
from datetime import date

from faker.providers import BaseProvider

from ... import utils


class Provider(BaseProvider):
    def raw_course_record(
        self, cohort_record: t.Optional[dict] = None, normalize_col_names: bool = False
    ) -> dict[str, object]:
        # use existing values where records overlap
        if cohort_record is not None:
            cr = cohort_record  # for more compact lines below
            student_guid = cr.get("student_guid", cr["Student GUID"])
            student_age = cr.get("student_age", cr["Student Age"])
            race = cr.get("race", cr["Race"])
            ethnicity = cr.get("ethnicity", cr["Ethnicity"])
            gender = cr.get("gender", cr["Gender"])
            institution_id = cr.get("institution_id", cr["Institution ID"])
            cohort = cr.get("cohort", cr["Cohort"])
            cohort_term = cr.get("cohort_term", cr["Cohort Term"])
            _has_enrollment_other_inst: bool = (
                cr.get(
                    "most_recent_bachelors_at_other_institution_state",
                    cr["Most Recent Bachelors at Other Institution STATE"],
                )
                is not None
            )
        else:
            student_guid = self.student_guid()
            student_age = self.student_age()
            race = self.race()
            ethnicity = self.ethnicity()
            gender = self.gender()
            institution_id = self.institution_id()
            cohort = self.cohort()
            cohort_term = self.cohort_term()
            _has_enrollment_other_inst: bool = self.generator.random.random() < 0.25  # type: ignore
        # derive a few more values, for self-consistency
        _min_academic_yr = int(cohort.split("-")[0])
        academic_year = self.academic_year(min_yr=_min_academic_yr)
        _min_course_yr = int(academic_year.split("-")[0])
        _max_course_yr = _min_course_yr + 1
        course_begin_date = self.course_begin_date(
            min_yr=_min_course_yr, max_yr=_max_course_yr
        )
        course_end_date = self.course_end_date(
            min_yr=_min_course_yr, max_yr=_max_course_yr
        )
        course_begin_date, course_end_date = sorted(
            [course_begin_date, course_end_date], reverse=False
        )  # ensure that begin date comes before end date
        core_course = self.core_course()
        core_course_type = self.core_course_type(core_course)
        num_credits_attempted = self.number_of_credits_attempted()
        num_credits_earned = self.number_of_credits_earned(
            max_value=num_credits_attempted
        )
        record = {
            "Student GUID": student_guid,
            "Student Age": student_age,
            "Race": race,
            "Ethnicity": ethnicity,
            "Gender": gender,
            "Institution ID": institution_id,
            "Cohort": cohort,
            "Cohort Term": cohort_term,
            "Academic Year": academic_year,
            "Academic Term": self.academic_term(),
            "Course Prefix": self.course_prefix(),
            "Course Number": self.course_number(),
            "Section ID": self.section_id(),
            "Course Name": self.course_name(),
            "Course CIP": self.course_cip(),
            "Course Type": self.course_type(),
            "Math or English Gateway": self.math_or_english_gateway(),
            "Co-requisite Course": self.co_requisite_course(),
            "Course Begin Date": course_begin_date,
            "Course End Date": course_end_date,
            "Grade": self.grade(),
            "Number of Credits Attempted": num_credits_attempted,
            "Number of Credits Earned": num_credits_earned,
            "Delivery Method": self.delivery_method(),
            "Core Course": core_course,
            "Core Course Type": core_course_type,
            "Core Competency Completed": self.core_competency_completed(),
            "Enrolled at Other Institution(s)": self.enrolled_at_other_institution_s(),
            "Credential Engine Identifier": self.credential_engine_identifier(),
            "Course Instructor Employment Status": self.course_instructor_employment_status(),
            "Course Instructor Rank": self.course_instructor_rank(),
            "Enrollment Record at Other Institution(s) STATE(s)": self.enrollment_record_at_other_institution_s_state_s(
                _has_enrollment_other_inst
            ),
            "Enrollment Record at Other Institution(s) CARNEGIE(s)": self.enrollment_record_at_other_institution_s_carnegie_s(
                _has_enrollment_other_inst
            ),
            "Enrollment Record at Other Institution(s) LOCALE(s)": self.enrollment_record_at_other_institution_s_locale_s(
                _has_enrollment_other_inst
            ),
        }
        if normalize_col_names:
            record = {
                utils.misc.convert_to_snake_case(key): val
                for key, val in record.items()
            }
        return record

    def student_guid(self) -> str:
        return self.numerify("#####!")  # type: ignore

    def institution_id(self) -> str:
        return self.numerify("#####!")  # type: ignore

    def student_age(self) -> str:
        return self.random_element(["20 AND YOUNGER", ">20 - 24", "OLDER THAN 24"])

    def race(self) -> str:
        return self.random_element(
            [
                "NONRESIDENT ALIEN",
                "AMERICAN INDIAN OR ALASKA NATIVE",
                "ASIAN",
                "BLACK OR AFRICAN AMERICAN",
                "NATIVE HAWAIIAN OR OTHER PACIFIC ISLANDER",
                "WHITE",
                "HISPANIC",
                "TWO OR MORE RACES",
                "UNKNOWN",
            ]
        )

    def ethnicity(self) -> str:
        return self.random_element(["H", "N", "UK"])

    def gender(self) -> str:
        return self.random_element(["M", "F", "P", "X", "UK"])

    def cohort(self, min_yr: int = 2010, max_yr: t.Optional[int] = None) -> str:
        start_dt = self.generator.date_between(start_date=date(min_yr, 1, 1))
        end_dt = start_dt.replace(year=start_dt.year + 1)
        return f"{start_dt:%Y}-{end_dt:%y}"

    def cohort_term(self) -> str:
        return self.random_element(["FALL", "WINTER", "SPRING", "SUMMER"])

    def academic_year(self, min_yr: int = 2010, max_yr: t.Optional[int] = None) -> str:
        start_dt = self.generator.date_between(
            start_date=date(min_yr, 1, 1),
            end_date=(date(max_yr, 1, 1) if max_yr else "today"),
        ).replace(day=1)
        end_dt = start_dt.replace(year=start_dt.year + 1)
        return f"{start_dt:%Y}-{end_dt:%y}"

    def academic_term(self) -> str:
        return self.random_element(["FALL", "WINTER", "SPRING", "SUMMER"])

    # TODO: more realistic course prefix and number?

    def course_prefix(self) -> str:
        return self.lexify("????").upper()

    def course_number(self) -> str:
        return self.numerify("##!")

    def section_id(self) -> str:
        return self.numerify("##!.#")

    def course_name(self) -> str:
        return " ".join(self.generator.words(nb=3, part_of_speech="noun")).upper()

    def course_cip(self) -> str:
        # TODO: make this six-digit CIP code more realistic
        return self.numerify("##.####")

    def course_type(self) -> str:
        return self.random_element(
            ["CU", "CG", "CC", "CD", "EL", "AB", "GE", "NC", "O"]
        )

    def math_or_english_gateway(self) -> str:
        return self.random_element(["M", "E", "NA"])

    def co_requisite_course(self) -> str:
        return self.random_element(["Y", "N"])

    def _course_date(self, min_yr: int = 2010, max_yr: t.Optional[int] = None) -> date:
        _end_date = date(max_yr, 1, 1) if max_yr is not None else "today"
        return self.generator.date_between(  # type: ignore
            start_date=date(min_yr, 1, 1), end_date=_end_date
        )

    def course_begin_date(
        self, min_yr: int = 2010, max_yr: t.Optional[int] = None
    ) -> date:
        return self._course_date(min_yr, max_yr)

    def course_end_date(
        self, min_yr: int = 2010, max_yr: t.Optional[int] = None
    ) -> date:
        return self._course_date(min_yr, max_yr)

    def grade(self) -> str:
        # TODO: use weighting for more realistic distribution?
        # find out if grades can ever be given as continuous values
        return self.random_element(
            ["0", "1", "2", "3", "4", "P", "F", "I", "W", "A", "M", "O"]
        )

    def _number_of_credits(
        self, min_value: float = 0.0, max_value: float = 20.0
    ) -> float:
        return self.generator.pyfloat(  # type: ignore
            min_value=min_value,
            max_value=max(max_value, min_value + 1e-3),
            right_digits=1,
        )

    def number_of_credits_attempted(self) -> float:
        return self._number_of_credits(min_value=1.0)

    def number_of_credits_earned(self, max_value: float = 20.0) -> float:
        return self._number_of_credits(max_value=max_value)

    def delivery_method(self) -> str:
        return self.random_element(["F", "O", "H"])

    def core_course(self) -> str:
        return self.random_element(["Y", "N"])

    def core_course_type(self, core_course: t.Optional[str] = None) -> t.Optional[str]:
        if core_course and core_course == "N":
            return None
        else:
            # TODO
            return "CORE COURSE TYPE"

    def core_competency_completed(self) -> str:
        return self.random_element(["Y", "N"])

    def enrolled_at_other_institution_s(self) -> str:
        return self.random_element(["Y", "N"])

    def credential_engine_identifier(self) -> t.Optional[str]:
        return None

    def course_instructor_employment_status(self) -> str:
        return self.random_element(["PT", "FT"])

    def course_instructor_rank(self) -> str:
        return self.random_element(["1", "2", "3", "4", "5", "6", "7"])

    def enrollment_record_at_other_institution_s_state_s(
        self, has_enrollment: bool = True
    ) -> t.Optional[str]:
        return (
            self.generator.state_abbr(include_freely_associated_states=False)
            if has_enrollment
            else None
        )  # type: ignore

    def enrollment_record_at_other_institution_s_carnegie_s(
        self, has_enrollment: bool = True
    ) -> t.Optional[str]:
        return (
            self.random_element(
                [
                    "High traditional",
                    "Mixed traditional/nontraditional",
                    "High nontraditional",
                    "Associate's Dominant",
                    "Mixed Baccalaureate/Associate's Colleges",
                    "Arts & sciences focus",
                    "Diverse fields",
                    "M1: Master's Colleges and Universities - Larger programs",
                    "M2: Master's Colleges and Universities – Medium programs",
                    "M3: Master's Colleges and Universities – Small programs",
                    "R1: Doctoral universities – very high research activity",
                    "R2: Doctoral universities – high research activity",
                    "D/PU: Doctoral/professional universities",
                ]
            )
            if has_enrollment
            else None
        )

    def enrollment_record_at_other_institution_s_locale_s(
        self, has_enrollment: bool = True
    ) -> t.Optional[str]:
        return (
            self.random_element(["URBAN", "SUBURB", "TOWN/RURAL"])
            if has_enrollment
            else None
        )



"""Task for publishing data that has been output by the inference task."""

import argparse
import logging

from student_success_tool.utils.gcs import publish_inference_output_files
from student_success_tool.utils.emails import send_inference_completion_email
from databricks.sdk import WorkspaceClient
from email.headerregistry import Address


# def save_file(
#     bucket: Bucket, src_volume_filepath: str, dest_bucket_pathname: str
# ) -> None:
#     """Save file from databricks volume to a GCP bucket path.

#     Args:
#       src_volume_filepath: The source filepath to the Databricks volume.
#       dest_bucket_pathname: The destination filepath in GCP.

#     Returns:
#       Nothing.
#     """
#     blob = bucket.blob(dest_bucket_pathname)
#     if blob.exists():
#         raise ValueError(dest_bucket_pathname + ": File already exists in bucket.")
#     blob.upload_from_filename(src_volume_filepath)


# def publish_inference_output_files(
#     db_workspace: str,
#     institution_name: str,
#     external_bucket_name: str,
#     sst_job_id: str,
#     approved: bool,
# ) -> str:
#     """Publish output files to bucket, with folder determined based on approved parameter,
#     and return the status of the job as a string.

#     Args:
#       db_workspace: The Databricks workspace to get files from.
#       institution_name: The Databricks institution name substring used in the schema (e.g. it would be
#         'uni_of_datakind' if the gold schema was 'uni_of_datakind_gold') -- this should match the
#         "Databricksified" name generated by the webapp.
#       external_bucket_name: The destination bucket in GCP.
#       sst_job_id: the job run id of this task.
#       approved: whether this file should be published in an approved state or not.

#     Returns:
#       The status string of the job run from Databricks.
#     """
#     # Destination: The output directory in the bucket that the files get moved to.
#     bucket_directory = f"{'approved' if approved else 'unapproved'}/{sst_job_id}"
#     # Source: The Databricks volume path we want to read the files from.
#     volume_path_top_level = f"/Volumes/{db_workspace}/{institution_name}_gold/gold_volume/inference_jobs/{sst_job_id}/ext"
#     volume_path_inference_folder = f"{volume_path_top_level}/inference_output"

#     storage_client = storage.Client()
#     bucket = storage_client.bucket(external_bucket_name)

#     # Full filepaths for the Databricks volume.
#     files_to_move = []
#     status_string = ""
#     # There will be other files in the inference_output folder generated by Databricks
#     # including time started/completed, and completion status.
#     for f in dbutils.fs.ls(volume_path_inference_folder):
#         filename = f.name
#         if filename.startswith("_"):
#             if not filename.startswith("_committed") and not filename.startswith(
#                 "_started"
#             ):
#                 status_string = filename
#         elif filename.endswith(".csv"):
#             files_to_move.append(f"{volume_path_inference_folder}/{filename}")
#     # The top level volume directory can include other files that are part
#     # of the published set of files. So we still need to iterate over these files. We only take .pngs from this directory
#     # we have to be selective with the extension type because there may be subfolders in this folder.
#     for f in dbutils.fs.ls(volume_path_top_level):
#         if f.name.endswith(".png"):
#             files_to_move.append(f"{volume_path_top_level}/{f.name}")

#     for f in files_to_move:
#         save_file(
#             bucket,
#             f,
#             f"{bucket_directory}/{os.path.basename(f)}",
#         )
#     return status_string


def main():
    """Main function."""
    logging.basicConfig(level=logging.INFO)
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--DB_workspace", required=True, help="Databricks workspace of the task."
    )
    parser.add_argument(
        "--databricks_institution_name",
        required=True,
        help="IThe Databricks institution name whose output we want to publish",
    )
    parser.add_argument(
        "--gcp_bucket_name",
        required=True,
        help="Name of the bucket we want to publish results to (in format [env]_[inst_id]).",
    )
    parser.add_argument(
        "--db_run_id",
        required=True,
        help="The job run id of the current pipeline execution.",
    )

    parser.add_argument(
        "--notification_email",
        required=True,
        help="User's email who triggered the inference run.",
    )
    args = parser.parse_args()
    w = WorkspaceClient()
    logging.info("Publishing files to GCP bucket")
    publish_inference_output_files(
        args.DB_workspace,
        args.databricks_institution_name,
        args.gcp_bucket_name,
        args.db_run_id,
        False,  # Set approved = False since this task will run immediately after inference.
    )
    username = w.dbutils.secrets.get(scope="sst", key="MANDRILL_USERNAME")
    sender_email = Address("Datakind Info", "help", "datakind.org")
    cc_email_list = ["education@datakind.org"]
    password = w.dbutils.secrets.get(scope="sst", key="MANDRILL_PASSWORD")
    logging.info("Sending email notification")
    send_inference_completion_email(
        sender_email, [args.notification_email], cc_email_list, username, password
    )


if __name__ == "__main__":
    main()



import os
import logging
import pandas as pd

from ..utils import utils

LOGGER = logging.getLogger(__name__)


def register_bias_sections(card, registry):
    """
    Registers the bias section of the model card with FNR group plots and
    subgroup disparity summaries. We focus on test data specifically for our reporting.

    This information is gathered from mlflow artifacts.
    """
    bias_levels = ["high", "moderate", "low"]
    bias_flags = {}
    group_disparities = {}

    def generate_description(
        group: str, subgroups: str, diff: str, stat_summary: str
    ) -> str:
        """
        Filters description for a given bias flag under 'bias_flags' directory in mlflow run
        artifacts. The FNR difference will be multiplied by 100 to represent a percentage and
        then rounded to an integer.

        Args:
            group: student group
            subgroups: student subgroups (e.g. "s1 vs. s2")
            diff: FNR (False Negative Rate) difference decimal (e.g. "0.1451")
            stat_summary: statistical summary with confidence interval & p-value information
        Returns:
            Description for a given bias flag
        """
        group_label = card.format.friendly_case(group)
        try:
            subgroup_1, subgroup_2 = [s.strip() for s in subgroups.split("vs")]
        except ValueError:
            LOGGER.warning(f"Could not parse subgroups for {group}: {subgroups}")
            return f"{card.format.bold('Could not parse subgroup comparison')}"

        sg1 = card.format.bold(card.format.italic(subgroup_1))
        sg2 = card.format.bold(card.format.italic(subgroup_2))
        percent_higher = card.format.bold(
            f"{int(round(float(diff) * 100))}% difference"
        )

        return f"- {sg1} students have a {percent_higher} in False Negative Rate (FNR) than {sg2} students. Statistical analysis indicates {stat_summary}."

    # Load bias flag CSVs and filter for test split
    for level in bias_levels:
        try:
            bias_path = f"bias_flags/{level}_bias_flags.csv"
            local_path = utils.download_artifact(
                run_id=card.run_id,
                local_folder=card.assets_folder,
                artifact_path=bias_path,
            )
            if not os.path.exists(local_path):
                LOGGER.warning(
                    f"{level} bias flags file does not exist. Bias evaluation likely has not run."
                )
                continue

            df = pd.read_csv(local_path)

            if df.empty or "split_name" not in df.columns:
                LOGGER.warning(
                    f"{level} bias flags file exists but has no data or missing 'split_name' column."
                )
                continue

            df = df[df["split_name"] == "test"]
            if df.empty:
                LOGGER.info(f"{level} bias flags file has no rows for test split.")
                continue

            # Group descriptions by group
            for _, row in df.iterrows():
                group = row["group"]
                desc = generate_description(
                    group,
                    row["subgroups"],
                    row["fnr_percentage_difference"],
                    row["type"],
                )
                group_disparities.setdefault(group, []).append(desc)
        except Exception as e:
            LOGGER.warning(
                f"Could not load {level} bias flags: [{type(e).__name__}] {str(e)}"
            )

    all_blocks = []

    for group_name, descriptions in group_disparities.items():
        normalized_name = group_name.lower().replace(" ", "_")
        plot_artifact_path = f"fnr_plots/test_{normalized_name}_fnr.png"

        try:
            plot_md = utils.download_artifact(
                run_id=card.run_id,
                local_folder=card.assets_folder,
                artifact_path=plot_artifact_path,
                description=f"False Negative Parity Rate for {group_name} on Test Data",
            )
        except Exception as e:
            LOGGER.warning(f"Could not load plot for {group_name}: {str(e)}")
            plot_md = (
                f"{card.format.bold(f'Unable to retrieve plot for {group_name}')}\n"
            )

        header = (
            f"{card.format.header_level(5)}{card.format.friendly_case(group_name)}\n\n"
        )
        text_block = "\n\n".join(descriptions)
        all_blocks.append(header + text_block + "\n\n" + plot_md)

    @registry.register("bias_groups_section")
    def bias_groups_section():
        """
        Returns just a filler text that can be used across any platform.
        """
        intro = f"{card.format.indent_level(1)}- Our assessment for FNR Parity was conducted across several student groups.\n"
        return intro

    @registry.register("bias_summary_section")
    def bias_summary_section():
        """
        Returns a markdown string containing the bias summary section of the model card.
        """
        if not all_blocks:
            LOGGER.warning(
                "No disparities found or bias evaluation artifacts missing. Skipping bias summary section."
            )
            return f"{card.format.italic('No statistically significant disparities were found on test dataset across groups.')}"

        section_header = (
            f"{card.format.header_level(4)}Disparities by Student Group\n\n"
        )
        return section_header + "\n\n".join(all_blocks)



import pytest
import os
from student_success_tool.dataio.read import from_toml_file
import re
from student_success_tool.modeling.inference import _get_mapped_feature_name


@pytest.fixture
def feature_table_data():
    project_root = os.getcwd()
    toml_path = os.path.join(
        project_root,
        "src",
        "student_success_tool",
        "assets",
        "pdp",
        "features_table.toml",
    )
    return from_toml_file(toml_path)


def test_toml_file_loads(feature_table_data):
    assert isinstance(feature_table_data, dict)
    assert len(feature_table_data) > 0, "The TOML file appears to be empty."


def test_all_features_have_name_and_desc(feature_table_data):
    for feature_id, entry in feature_table_data.items():
        assert isinstance(entry, dict), f"Entry for {feature_id} should be a dict."
        assert "name" in entry and entry["name"].strip(), (
            f"'name' missing or empty in feature: {feature_id}"
        )
        assert "short_desc" in entry and entry["short_desc"].strip(), (
            f"'short_desc' missing or empty in feature: {feature_id}"
        )
        assert "long_desc" in entry and entry["long_desc"].strip(), (
            f"'long_desc' missing or empty in feature: {feature_id}"
        )


# Add regex test cases to this as needed! This helps us make sure
# that are regex patterns work *as intended* in our features table
VALID_FEATURE_NAMES = [
    "took_course_id_eng_101",
    "cummax_took_course_id_eng_101",
    "cummax_in_12_creds_took_course_id_eng_101",
    "took_course_subject_area_51",
    "cummax_took_course_subject_area_51",
    "cummax_in_12_creds_took_course_subject_area_51",
    "took_course_id_english_composition_and_writing_098",
    "cummax_took_course_id_english_composition_and_writing_098",
    "cummax_in_12_creds_took_course_id_english_composition_and_writing_098",
    "took_course_id_english_composition_and_writing_101_1",
    "cummax_took_course_id_english_composition_and_writing_101_1",
    "cummax_in_12_creds_took_course_id_english_composition_and_writing_101_1",
    "num_courses_course_subject_area_51",
    "cumfrac_num_courses_course_subject_area_51",
    "frac_courses_course_subject_area_51",
    "num_courses_course_id_eng_101",
    "cumfrac_num_courses_course_id_eng_101",
    "frac_courses_course_id_eng_101",
    "took_course_id_english1010",
    "cummax_took_course_id_english1010",
    "cummax_in_12_creds_took_course_id_english1010",
    "took_course_id_mathematics_science1210",
    "cummax_took_course_id_mathematics_science1210",
    "cummax_in_12_creds_took_course_id_mathematics_science1210",
    "took_course_id_computer_science_1050",
    "cummax_took_course_id_computer_science_1050",
    "cummax_in_12_creds_took_course_id_communication_studies_1010",
    "cummax_in_12_creds_took_course_id_computer_science_1050",
    "num_courses_course_id_mathematics_science1210",
    "num_courses_course_id_english1010",
    "num_courses_course_id_english_composition_and_writing_098",
    "num_courses_course_id_english_composition_and_writing_101_1",
    "cumfrac_num_courses_course_id_english_1020",
    "cumfrac_num_courses_course_id_mathematics_science1210",
    "cumfrac_num_courses_course_id_english1010",
    "cumfrac_num_courses_course_id_philosophy_1010",
    "cumfrac_num_courses_course_id_english_composition_and_writing_098",
    "cumfrac_num_courses_course_id_english_composition_and_writing_101_1",
    "frac_courses_course_id_mathematics_science1210",
    "frac_courses_course_id_english1010",
    "frac_courses_course_id_english_composition_and_writing_098",
    "frac_courses_course_id_english_composition_and_writing_101_1",
]


@pytest.mark.parametrize("feature_name", VALID_FEATURE_NAMES)
def test_feature_maps_to_named_entry(feature_name, feature_table_data):
    """Ensure each feature maps to a named entry in the feature table using the production mapping logic."""
    mapped = _get_mapped_feature_name(feature_name, feature_table_data)

    # If it mapped to something, it must not be the identity (fallback case)
    assert mapped != feature_name, (
        f"Feature '{feature_name}' was not mapped correctly using _get_mapped_feature_name."
    )

    # Ensure that the result is a formatted name string
    assert isinstance(mapped, str) and mapped.strip(), (
        f"Mapped name for '{feature_name}' is empty or invalid: {mapped}"
    )

    # check for each regex pattern the name matches as well
    if re.fullmatch(
        r"^cummax_in_12_creds_took_course_subject_area_(.*)$", feature_name
    ):
        subject_area = re.search(
            r"^cummax_in_12_creds_took_course_subject_area_(.*)$", feature_name
        ).group(1)
        assert (
            mapped
            == f"took course in subject area '{subject_area}' within student's first 12 credits"
        )
    elif re.fullmatch(r"^cummax_took_course_subject_area_(.*)$", feature_name):
        subject_area = re.search(
            r"^cummax_took_course_subject_area_(.*)$", feature_name
        ).group(1)
        assert mapped == f"course in subject area '{subject_area}' taken so far"
    elif re.fullmatch(r"^took_course_subject_area_(.*)$", feature_name):
        subject_area = re.search(
            r"^took_course_subject_area_(.*)$", feature_name
        ).group(1)
        assert mapped == f"courses taken in subject area '{subject_area}' this term"
    elif re.fullmatch(r"^frac_courses_course_subject_area_(.*)$", feature_name):
        subject_area = re.search(
            r"^frac_courses_course_subject_area_(.*)$", feature_name
        ).group(1)
        assert (
            mapped
            == f"fraction of courses taken in subject area {subject_area} this term"
        )
    elif re.fullmatch(r"^cumfrac_num_courses_course_subject_area_(.*)$", feature_name):
        subject_area = re.search(
            r"^cumfrac_num_courses_course_subject_area_(.*)$", feature_name
        ).group(1)
        assert (
            mapped == f"fraction of courses taken in subject area {subject_area} so far"
        )
    elif re.fullmatch(r"^num_courses_course_subject_area_(.*)$", feature_name):
        subject_area = re.search(
            r"^num_courses_course_subject_area_(.*)$", feature_name
        ).group(1)
        assert (
            mapped
            == f"number of courses taken in subject area {subject_area} this term"
        )
    elif re.fullmatch(r"^cummax_in_12_creds_took_course_id_(.*)$", feature_name):
        subject_area = re.search(
            r"^cummax_in_12_creds_took_course_id_(.*)$", feature_name
        ).group(1)
        assert (
            mapped
            == f"course with id '{subject_area}' taken within student's first 12 credits"
        )
    elif re.fullmatch(r"^cummax_took_course_id_(.*)$", feature_name):
        subject_area = re.search(r"^cummax_took_course_id_(.*)$", feature_name).group(1)
        assert mapped == f"course with id '{subject_area}' taken so far"
    elif re.fullmatch(r"^took_course_id_(.*)$", feature_name):
        subject_area = re.search(r"^took_course_id_(.*)$", feature_name).group(1)
        assert mapped == f"course with id '{subject_area}' taken this term"
    elif re.fullmatch(r"^frac_courses_course_id_(.*)$", feature_name):
        subject_area = re.search(r"^frac_courses_course_id_(.*)$", feature_name).group(
            1
        )
        assert mapped == f"fraction of times course '{subject_area}' taken this term"
    elif re.fullmatch(r"^cumfrac_num_courses_course_id_(.*)$", feature_name):
        subject_area = re.search(
            r"^cumfrac_num_courses_course_id_(.*)$", feature_name
        ).group(1)
        assert mapped == f"fraction of times course '{subject_area}' taken so far"
    elif re.fullmatch(r"^num_courses_course_id_(.*)$", feature_name):
        subject_area = re.search(r"^num_courses_course_id_(.*)$", feature_name).group(1)
        assert mapped == f"number of times course '{subject_area}' taken this term"

    # Ensure only one regex pattern matches this feature
    matching_patterns = [
        pattern for pattern in feature_table_data if re.fullmatch(pattern, feature_name)
    ]
    assert len(matching_patterns) == 1, (
        f"Feature '{feature_name}' matches {len(matching_patterns)} patterns: {matching_patterns}. "
        "Feature should match exactly one pattern."
    )



# Databricks notebook source
# MAGIC %md
# MAGIC # SST Inference Validation
# MAGIC
# MAGIC The last step is our human-in-the-loop process before we send files back to institutions. It is a critical step to ensure that of our process to ensure high quality & consistency.
# MAGIC
# MAGIC ### References
# MAGIC
# MAGIC - [Data science product components (Confluence doc)](https://datakind.atlassian.net/wiki/spaces/TT/pages/237862913/Data+science+product+components+the+modeling+process)
# MAGIC - [Databricks runtimes release notes](https://docs.databricks.com/en/release-notes/runtime/index.html)
# MAGIC

# COMMAND ----------

# MAGIC %sh python --version

# COMMAND ----------

# WARNING: AutoML/mlflow expect particular packages with version constraints
# that directly conflicts with dependencies in our SST repo. As a temporary fix,
# we need to manually install a certain version of pandas and scikit-learn in order
# for our models to load and run properly.

# %pip install "student-success-tool==0.3.7"
# %restart_python

# COMMAND ----------

import os
import sys
import logging
from databricks.connect import DatabricksSession

import matplotlib.pyplot as plt

from student_success_tool import dataio, configs

if "DATABRICKS_RUNTIME_VERSION" in os.environ:
    sys.path.insert(1, "../")

run_type = "predict"  # always predict

try:
    spark = DatabricksSession.builder.getOrCreate()
except Exception:
    logging.warning("unable to create spark session; are you in a Databricks runtime?")
    pass

# COMMAND ----------

# load config file
cfg = dataio.read_config(
    "./config-TEMPLATE.toml", schema=configs.custom.CustomProjectConfig
)
cfg

# COMMAND ----------

gold_prediction_table_name = getattr(
    cfg.datasets.gold["advisor_output"], f"{run_type}_table_path"
)
print(gold_prediction_table_name)

predict_df = dataio.read.from_delta_table(
    gold_prediction_table_name,
    spark_session=spark,
)
predict_df

# COMMAND ----------

# Extract student id from config
student_id_col = cfg.student_id_col

# Assert that final output has rows
assert predict_df.shape[0] > 0, "Final output has no rows."

# Report the number of students
num_students = len(predict_df[student_id_col])
print(f"Number of students: {num_students}")

# Assert that all STUDENT_ID values are unique
assert predict_df[student_id_col].is_unique, (
    "The student ID column contains duplicate values."
)

# COMMAND ----------

# MAGIC %md
# MAGIC ## Validate distribution in comparison to training

# COMMAND ----------

predict_df["support_score"].hist(bins=50)
plt.title("Distribution of Support Scores on Inference Dataset")
plt.ylabel("# of Students")
plt.xlabel("support_score")

# COMMAND ----------

predict_df["support_score"].value_counts(normalize=True)

# COMMAND ----------

# MAGIC %md
# MAGIC ## Look at Top Indicator Value Counts across all Students
# MAGIC If there's not many top features across the student population, it could be the sign of invalid predictions or an unhealthy model.

# COMMAND ----------

predict_df[[f"feature_{i}_name" for i in range(1, 6)]].describe()

# COMMAND ----------

# MAGIC %md
# MAGIC ## Validate that SHAP values are decreasing in magnitude for each student

# COMMAND ----------

assert all(
    all(
        abs(predict_df.loc[i, f"feature_{j}_importance"])
        >= abs(predict_df.loc[i, f"feature_{j + 1}_importance"])
        for j in range(1, 5)
    )
    for i in range(len(predict_df))
), (
    "Final output has invalid SHAP values across top ranked features for one or more students."
)

# COMMAND ----------

# MAGIC %md
# MAGIC ## Check whether there are feature values that are 'nan'

# COMMAND ----------

predict_df.loc[
    (predict_df[[f"feature_{i}_value" for i in range(1, 6)]] == "nan").any(axis=1)
]

# COMMAND ----------

# MAGIC %md
# MAGIC ## Validate examples for the most, median, and least risk students

# COMMAND ----------

# Maximum Risk Student
max_student = predict_df[
    predict_df["support_score"] == predict_df["support_score"].max()
]
max_student

# COMMAND ----------

# Minimum Risk Student
min_student = predict_df[
    predict_df["support_score"] == predict_df["support_score"].min()
]
min_student

# COMMAND ----------

# Median Risk Student
median_value = predict_df["support_score"].median()
median_student = predict_df.loc[
    (predict_df["support_score"] - median_value).abs().nsmallest(1).index
]
median_student



import pytest
from unittest.mock import MagicMock
from student_success_tool.reporting.sections.registry import SectionRegistry
from student_success_tool.reporting.sections.pdp import (
    attribute_sections as pdp_attribute_sections,
)
from student_success_tool.reporting.utils.formatting import Formatting


@pytest.fixture
def mock_card():
    card = MagicMock()
    card.format = Formatting()
    return card


@pytest.mark.parametrize(
    "outcome_type, time_limits, extra_config, expected_snippet",
    [
        (
            "retention",
            {},
            {},
            "The model predicts the likelihood of non-retention into the student's second academic year based on student, course, and academic data.",
        ),
        (
            "graduation",
            {"FULL-TIME": (2.0, "year"), "PART-TIME": (3.0, "year")},
            {},
            "The model predicts the likelihood of not graduating on time within 2 years for full-time students, and within 3 years for part-time students, based on student, course, and academic data.",
        ),
        (
            "credits_earned",
            {"FULL-TIME": (1.5, "year")},
            {"min_num_credits": 45},
            "The model predicts the likelihood of not earning 45 credits within 1.5 years for full-time students, based on student, course, and academic data.",
        ),
    ],
)
def test_outcome_variants(
    mock_card, outcome_type, time_limits, extra_config, expected_snippet
):
    mock_card.cfg.preprocessing.target.type_ = outcome_type
    mock_card.cfg.preprocessing.selection.intensity_time_limits = time_limits

    # Patching checkpoint since we are testing outcome
    mock_card.cfg.preprocessing.checkpoint.type_ = "nth"
    mock_card.cfg.preprocessing.checkpoint.n = 1
    mock_card.cfg.preprocessing.checkpoint.exclude_pre_cohort_terms = False
    mock_card.cfg.preprocessing.checkpoint.exclude_non_core_terms = False
    mock_card.cfg.preprocessing.checkpoint.valid_enrollment_year = None

    if outcome_type == "credits_earned" and "min_num_credits" in extra_config:
        mock_card.cfg.preprocessing.target.min_num_credits = extra_config[
            "min_num_credits"
        ]

    registry = SectionRegistry()
    pdp_attribute_sections.register_attribute_sections(mock_card, registry)
    rendered = registry.render_all()

    assert expected_snippet in rendered["outcome_section"]


def test_target_population_section(mock_card):
    mock_card.cfg.preprocessing.selection.student_criteria = {
        "degree": ["bachelor's", "associate's"],
        "status": "full-time",
    }

    # Patching checkpoint since we are testing target population
    mock_card.cfg.preprocessing.checkpoint.type_ = "nth"
    mock_card.cfg.preprocessing.checkpoint.n = 1
    mock_card.cfg.preprocessing.checkpoint.exclude_pre_cohort_terms = False
    mock_card.cfg.preprocessing.checkpoint.exclude_non_core_terms = False
    mock_card.cfg.preprocessing.checkpoint.valid_enrollment_year = None

    registry = SectionRegistry()
    pdp_attribute_sections.register_attribute_sections(mock_card, registry)
    rendered = registry.render_all()

    assert "Degree" in rendered["target_population_section"]
    assert "- Bachelor's" in rendered["target_population_section"]
    assert "- Associate's" in rendered["target_population_section"]
    assert "Status" in rendered["target_population_section"]
    assert "- Full-Time" in rendered["target_population_section"]


@pytest.mark.parametrize(
    "checkpoint_type,n,exclude_pre,exclude_non_core,valid_year,expected_output",
    [
        # Basic nth
        (
            "nth",
            1,
            True,
            True,
            None,
            "The model makes this prediction when the student has completed their 2nd term.",
        ),
        # Include pre-cohort terms only
        (
            "nth",
            3,
            False,
            True,
            None,
            "The model makes this prediction when the student has completed their 4th term including pre-cohort terms.",
        ),
        # Include non-core terms only
        (
            "nth",
            3,
            True,
            False,
            None,
            "The model makes this prediction when the student has completed their 4th term including non-core terms.",
        ),
        # Include both
        (
            "nth",
            3,
            False,
            False,
            None,
            "The model makes this prediction when the student has completed their 4th term including pre-cohort terms and non-core terms.",
        ),
        # Include both + valid enrollment year
        (
            "nth",
            -1,
            False,
            False,
            1,
            "The model makes this prediction when the student has completed their last term including pre-cohort terms and non-core terms, provided the term occurred in their 1st year of enrollment.",
        ),
        # Only valid enrollment year
        (
            "nth",
            6,
            True,
            True,
            3,
            "The model makes this prediction when the student has completed their 7th term, provided the term occurred in their 3rd year of enrollment.",
        ),
        (
            "first",
            0,
            False,
            False,
            None,
            "The model makes this prediction when the student has completed their first term.",
        ),
        (
            "last",
            -1,
            False,
            False,
            None,
            "The model makes this prediction when the student has completed their last term.",
        ),
        (
            "first_at_num_credits_earned",
            None,
            None,
            None,
            None,
            "The model makes this prediction when the student has earned 30 credits.",
        ),
        (
            "last_in_enrollment_year",
            None,
            None,
            None,
            None,
            "The model makes this prediction when the student has completed their 2nd year of enrollment.",
        ),
        (
            "first_within_cohort",
            None,
            None,
            None,
            None,
            "The model makes this prediction when the student has completed their first term within their cohort.",
        ),
    ],
)
def test_checkpoint_variants(
    mock_card,
    checkpoint_type,
    n,
    exclude_pre,
    exclude_non_core,
    valid_year,
    expected_output,
):
    mock_card.cfg.preprocessing.checkpoint.type_ = checkpoint_type

    if checkpoint_type == "nth":
        mock_card.cfg.preprocessing.checkpoint.n = n
        mock_card.cfg.preprocessing.checkpoint.exclude_pre_cohort_terms = exclude_pre
        mock_card.cfg.preprocessing.checkpoint.exclude_non_core_terms = exclude_non_core
        mock_card.cfg.preprocessing.checkpoint.valid_enrollment_year = valid_year
    elif checkpoint_type == "first_at_num_credits_earned":
        mock_card.cfg.preprocessing.checkpoint.min_num_credits = 30
    elif checkpoint_type == "last_in_enrollment_year":
        mock_card.cfg.preprocessing.checkpoint.enrollment_year = 2

    registry = SectionRegistry()
    pdp_attribute_sections.register_attribute_sections(mock_card, registry)
    rendered = registry.render_all()

    assert rendered["checkpoint_section"].strip().endswith(expected_output)


def test_checkpoint_unknown_type_raises_error(mock_card):
    mock_card.cfg.preprocessing.checkpoint.type_ = "unknown_metric"

    registry = SectionRegistry()
    pdp_attribute_sections.register_attribute_sections(mock_card, registry)

    with pytest.raises(ValueError, match="Unknown checkpoint type: unknown_metric"):
        registry.render_all()



import logging
import os
import shutil
import typing as t
import uuid
from collections.abc import Sequence
from typing import Optional, List
import matplotlib.colors as mcolors
import matplotlib.figure
import matplotlib.pyplot as plt
import mlflow
import mlflow.artifacts
import numpy as np
import pandas as pd
import seaborn as sns
import shap
import sklearn.inspection
import sklearn.metrics
import sklearn.utils
from sklearn.calibration import calibration_curve
from sklearn.preprocessing import MinMaxScaler
from statsmodels.nonparametric.smoothers_lowess import lowess

try:
    from IPython.display import display
except ImportError:
    display = t.cast(t.Callable[..., t.Any], print)

LOGGER = logging.getLogger(__name__)

# TODO: eventually we should use the custom_style.mplstyle colors, but currently
# the color palette does not have distinct enough colors for calibration by group
# where there are a lot of subgroups
PALETTE = sns.color_palette("Paired")

PosLabelType = t.Optional[int | float | bool | str]


def extract_training_data_from_model(
    automl_experiment_id: str, data_runname: str = "Training Data Storage and Analysis"
) -> pd.DataFrame:
    """
    Read training data from a model into a pandas DataFrame. This allows us to run more
    evaluations of the model, ensuring that we are using the same train/test/validation split

    Args:
        automl_experiment_id: Experiment ID of the AutoML experiment
        data_runname: The runName tag designating where there training data is stored

    Returns:
        The data used for training a model, with train/test/validation flags
    """
    run_df = mlflow.search_runs(
        experiment_ids=[automl_experiment_id], output_format="pandas"
    )
    assert isinstance(run_df, pd.DataFrame)  # type guard
    data_run_id = run_df[run_df["tags.mlflow.runName"] == data_runname]["run_id"].item()

    # Create temp directory to download input data from MLflow
    input_temp_dir = os.path.join(
        os.environ["SPARK_LOCAL_DIRS"], "tmp", str(uuid.uuid4())[:8]
    )
    os.makedirs(input_temp_dir)

    # Download the artifact and read it into a pandas DataFrame
    input_data_path = mlflow.artifacts.download_artifacts(
        run_id=data_run_id, artifact_path="data", dst_path=input_temp_dir
    )
    df_loaded = pd.read_parquet(os.path.join(input_data_path, "training_data"))
    # Delete the temp data
    shutil.rmtree(input_temp_dir)

    return df_loaded


def evaluate_performance(
    df_pred: pd.DataFrame,
    *,
    pos_label: PosLabelType,
    split_col: str = "split",
    target_col: str = "target",
    pred_col: str = "pred",
    pred_prob_col: str = "pred_prob",
) -> None:
    """
    Evaluates and logs model performance for each data split. Generates
    histogram and calibration plots.

    Args:
        df_pred: DataFrame containing prediction results with a column for splits.
        split_col: Column name indicating split column ("train", "test", or "val").
        target_col: Column name for the target variable.
        pred_prob_col: Column name for predicted probabilities.
        pos_label: Positive class label.
    """
    calibration_dir = "calibration"
    preds_dir = "preds"
    metrics_dir = "metrics"
    metrics_records = []

    for split_name, split_data in df_pred.groupby(split_col):
        LOGGER.info("Evaluating model performance for '%s' split", split_name)
        split_data.to_csv(f"/tmp/{split_name}_preds.csv", header=True, index=False)
        mlflow.log_artifact(
            local_path=f"/tmp/{split_name}_preds.csv",
            artifact_path=preds_dir,
        )
        hist_fig, cal_fig = create_evaluation_plots(
            split_data,
            pred_prob_col,
            target_col,
            pos_label,
            split_name,
        )
        mlflow.log_figure(hist_fig, os.path.join(preds_dir, f"{split_name}_hist.png"))
        mlflow.log_figure(
            cal_fig, os.path.join(calibration_dir, f"{split_name}_calibration.png")
        )
        # Closes all matplotlib figures in console to free memory
        plt.close("all")

        # Compute performance metrics by split (returns a dict)
        perf_metrics_raw = compute_classification_perf_metrics(
            targets=split_data[target_col],
            preds=split_data[pred_col],
            pred_probs=split_data[pred_prob_col],
            pos_label=pos_label,
        )

        perf_metrics = format_perf_metrics(perf_metrics_raw)
        perf_split_col = f"Dataset {split_col.capitalize()}"
        split_map = {"test": "Test", "train": "Training", "validate": "Validation"}
        perf_metrics[perf_split_col] = split_map.get(split_name)
        metrics_records.append(perf_metrics)

    # Convert to DataFrame for display or saving
    metrics_df = pd.DataFrame(metrics_records).set_index(perf_split_col)
    metrics_df.to_csv("/tmp/performance_across_splits.csv")
    mlflow.log_artifact("/tmp/performance_across_splits.csv", artifact_path=metrics_dir)
    LOGGER.info("Creating summary of performance metrics across splits")


def format_perf_metrics(perf_metrics_raw: dict) -> dict:
    """
    Formats performance metrics from raw metrics. This is for a model cards
    table that presents performance by splits.

    Args:
        perf_metrics_raw: Dictionary containing raw performance metrics.

    Returns:
        Dictionary summarizing the formatted performance metrics.
    """
    return {
        "Number of Samples": int(perf_metrics_raw["num_samples"]),
        "Number of Positive Samples": int(perf_metrics_raw["num_positives"]),
        "Actual Target Prevalence": round(
            float(perf_metrics_raw["true_positive_prevalence"]), 2
        ),
        "Predicted Target Prevalence": round(
            float(perf_metrics_raw["pred_positive_prevalence"]), 2
        ),
        "Accuracy": round(float(perf_metrics_raw["accuracy"]), 2),
        "Precision": round(float(perf_metrics_raw["precision"]), 2),
        "Recall": round(float(perf_metrics_raw["recall"]), 2),
        "F1 Score": round(float(perf_metrics_raw["f1_score"]), 2),
        "Log Loss": round(float(perf_metrics_raw["log_loss"]), 2),
    }


def get_top_runs(
    experiment_id: str,
    optimization_metrics: t.Union[str, list[str]],
    topn_runs_included: int = 1,
) -> dict[str, str]:
    """
    Retrieve top runs from an MLflow experiment using a balanced score over multiple metrics.

    Parameters:
        experiment_id: MLflow experiment ID.
        optimization_metrics: Single metric or list of metrics to sort by.
        topn_runs_included: Number of top runs to return.

    Returns:
        A dictionary mapping run_name to run_id.
    """
    if isinstance(optimization_metrics, str):
        optimization_metrics = [optimization_metrics]

    # Prepare column names and directions
    sort_by = [f"metrics.{m}" for m in optimization_metrics]
    directions = infer_directions(optimization_metrics)

    # Fetch all runs
    runs = mlflow.search_runs(experiment_ids=[experiment_id], output_format="pandas")
    assert isinstance(runs, pd.DataFrame)

    # Clean and filter
    valid_runs = runs.dropna(subset=sort_by).copy()
    for col in sort_by:
        valid_runs[col] = valid_runs[col].astype(float)

    # Score and sort
    valid_runs["balanced_score"] = compute_balanced_score(
        valid_runs, sort_by, directions
    )
    sorted_runs = valid_runs.sort_values("balanced_score")

    # Display viz of top runs
    display_cols = ["tags.mlflow.runName", "run_id"] + sort_by + ["balanced_score"]

    print("Inferred directions:")
    for metric, direction in zip(optimization_metrics, directions):
        print(f"  • {metric}: {'minimize' if direction == 'asc' else 'maximize'}")

    print(f"\n Top {topn_runs_included} runs (by balanced score):")
    display(sorted_runs[display_cols].head(topn_runs_included))

    # Return as {run_name: run_id}
    top_runs = sorted_runs.head(topn_runs_included)
    return {
        row.get("tags.mlflow.runName", f"Unnamed_{row['run_id']}"): row["run_id"]
        for _, row in top_runs.iterrows()
    }


def infer_directions(metrics: list[str]) -> list[str]:
    """
    When getting top runs, we can infer optimization direction
    for each metric: 'asc' (minimize) or 'desc' (maximize).

    Parameters:
        metrics: List of metric names.

    Returns:
        List of directions, one for each metric.
    """
    minimize_keywords = ("loss", "error", "log_loss", "mae", "mse")
    return [
        "asc" if any(metric.lower().endswith(k) for k in minimize_keywords) else "desc"
        for metric in metrics
    ]


def compute_balanced_score(
    df: pd.DataFrame, metrics: list[str], directions: list[str]
) -> pd.Series:
    """
    When sorting through top runs, we need to average normalize our metrics in order to
    properly find the best overall runs across all given metrics. We balance ALL metrics
    equally. We then compute a balanced average score (lower is better).

    TODO: Provide a priority list of metrics: higher recall, then ROC AUC, then log loss,
    for example.

    Parameters:
        df: Pandas DataFrame containing metrics
        metrics: List of metric names
        directions: List of directions for each metric

    Returns:
        Balanced score for each row in the DataFrame
    """
    scaler = MinMaxScaler()
    data = df[metrics].copy()
    normalized = scaler.fit_transform(data)

    # Flip maximization metrics so that lower = better for everything
    for i, direction in enumerate(directions):
        if direction == "desc":
            normalized[:, i] = 1 - normalized[:, i]

    return pd.Series(normalized.mean(axis=1), index=df.index)


def compute_classification_perf_metrics(
    targets: pd.Series,
    preds: pd.Series,
    pred_probs: pd.Series,
    *,
    pos_label: PosLabelType,
    sample_weights: t.Optional[pd.Series] = None,
) -> dict[str, object]:
    """
    Compute a variety of useful metrics for evaluating the performance of a classifier.

    Args:
        targets: True target values that classifier model was trained to predict.
        preds: Binary predictions output by classifier model.
        pred_probs: Prediction probabilities output by classifier model,
            for the positive class only.
        pos_label: Value of the positive class (aka "label").
        sample_weights: Sample weights for the examples corresponding to targets/preds,
            if classifier model was trained on weighted samples.
    """
    assert isinstance(pos_label, (int, float, bool, str))  # type guard
    precision, recall, f1_score, _ = sklearn.metrics.precision_recall_fscore_support(
        targets,
        preds,
        beta=1,
        average="binary",
        pos_label=pos_label,  # type: ignore
        sample_weight=sample_weights,
        zero_division=np.nan,  # type: ignore
    )
    result = {
        "num_samples": len(targets),
        "num_positives": targets.eq(pos_label).sum(),
        "true_positive_prevalence": targets.eq(pos_label).mean(),
        "pred_positive_prevalence": preds.eq(pos_label).mean(),
        "accuracy": sklearn.metrics.accuracy_score(
            targets, preds, sample_weight=sample_weights
        ),
        "precision": precision,
        "recall": recall,
        "f1_score": f1_score,
        "log_loss": sklearn.metrics.log_loss(
            targets, pred_probs, sample_weight=sample_weights, labels=[False, True]
        ),
        # TODO: add this back in, but handle errors for small N case
        # "roc_auc": sklearn.metrics.roc_auc_score(
        #     targets, pred_probs, sample_weight=sample_weights, labels=[False, True]
        # ),
    }
    return result


def get_sensitivity_of_top_q_pctl_thresh(
    y_true: pd.Series | Sequence,
    risk_score: pd.Series | Sequence,
    q: float,
    pos_label: PosLabelType,
) -> float:
    """
    Report sensitivity (AKA recall score) using some percentile threshold.

    Calculation:
        number of true positivies / (number of true positives + number of false negatives)
        OR
        number of true positives / total number of actual true class

    Args:
        y_true: true outcome class, of length n_samples
        risk_score: predicted risk scores, of length n_samples
        q: probability for the quantiles to compute
        pos_label: label identifying the positive class in y_true
    """
    if not isinstance(y_true, pd.Series):
        y_true = pd.Series(y_true)
    if not isinstance(risk_score, pd.Series):
        risk_score = pd.Series(risk_score)

    prob_thresh = np.quantile(risk_score, q)
    high_risk = risk_score >= prob_thresh

    # convert actual outcome to booleans to match high_risk array
    y_true = y_true.apply(lambda x: True if x == pos_label else False)

    result = sklearn.metrics.recall_score(y_true, high_risk)
    assert isinstance(result, float)
    return result


def compare_trained_models(
    automl_experiment_id: str, automl_metric: str
) -> tuple[pd.DataFrame, str]:
    """
    Retrieve, aggregate and sort performance data for models trained in a specified AutoML experiment.
    The validation dataset is used to tune hyperparameters. Metrics on the validation dataset are used to rank models, so we also use this metric to compare across models.

    Args:
        automl_experiment_id: Experiment ID of the AutoML experiment
        automl_metric: Chosen AutoML optimization metric

    Returns:
        DataFrame containing model types and highest scores for the given metric.
    """
    runs = mlflow.search_runs(
        experiment_ids=[automl_experiment_id], output_format="pandas"
    )
    assert isinstance(runs, pd.DataFrame)  # type guard
    metric = str.lower(automl_metric)
    metric_score_column = (
        f"metrics.val_{metric}"
        if metric == "log_loss"
        else f"metrics.val_{metric}_score"
    )

    runs = runs[["tags.model_type", metric_score_column]].dropna()

    ascending = metric == "log_loss"
    df_sorted = (
        runs.groupby("tags.model_type", as_index=False)
        .agg({metric_score_column: ("min" if ascending else "max")})
        .sort_values(by=metric_score_column, ascending=ascending)
    )
    return df_sorted, metric_score_column


def compute_feature_permutation_importance(
    model: sklearn.base.BaseEstimator,
    features: pd.DataFrame,
    target: pd.Series,
    *,
    n_repeats: int = 5,
    scoring: t.Optional[str | t.Callable] = None,
    sample_weight: t.Optional[np.ndarray] = None,
    random_state: t.Optional[int] = None,
) -> sklearn.utils.Bunch:
    """
    Compute model features' permutation importance.

    Args:
        model
        features
        target
        n_repeats
        scoring
        sample_weight
        random_state

    References:
        - https://scikit-learn.org/stable/modules/generated/sklearn.inspection.permutation_importance.html
        - https://scikit-learn.org/stable/modules/model_evaluation.html#callable-scorers

    See Also:
        - :func:`plot_feature_permutation_importances()`
    """
    result = sklearn.inspection.permutation_importance(
        model,
        features,
        target,
        scoring=scoring,
        n_repeats=n_repeats,
        sample_weight=sample_weight,
        random_state=random_state,
    )
    assert isinstance(result, sklearn.utils.Bunch)  # type guard
    return result


def log_confusion_matrix(
    institution_id: str,
    *,
    automl_run_id: str,
    catalog: str = "staging_sst_01",
) -> None:
    """
    Register an mlflow model according to one of their various recommended approaches.

    Args:
        institution_id
        automl_run_id
        catalog
    """
    try:
        from databricks.connect import DatabricksSession

        spark = DatabricksSession.builder.getOrCreate()
    except Exception:
        print("Databricks Connect failed. Falling back to local Spark.")
        from pyspark.sql import SparkSession

        spark = (
            SparkSession.builder.master("local[*]").appName("Fallback").getOrCreate()
        )

    confusion_matrix_table_path = (
        f"{catalog}.{institution_id}_silver.training_{automl_run_id}_confusion_matrix"
    )

    def safe_div(numerator, denominator):
        return numerator / denominator if denominator else 0.0

    if mlflow.active_run() is None:
        raise RuntimeError(
            "No active MLflow run. Please call this within an MLflow run context."
        )

    try:
        run = mlflow.get_run(automl_run_id)
        required_metrics = [
            "test_true_positives",
            "test_true_negatives",
            "test_false_positives",
            "test_false_negatives",
        ]
        metrics = {m: run.data.metrics.get(m) for m in required_metrics}

        if any(v is None for v in metrics.values()):
            raise ValueError(
                f"Missing one or more required metrics in run {automl_run_id}: {metrics}"
            )

        tp, tn, fp, fn = (
            metrics["test_true_positives"],
            metrics["test_true_negatives"],
            metrics["test_false_positives"],
            metrics["test_false_negatives"],
        )

        tn_percentage = safe_div(tn, tn + fp)
        tp_percentage = safe_div(tp, tp + fn)
        fp_percentage = safe_div(fp, fp + tn)
        fn_percentage = safe_div(fn, fn + tp)

        confusion_matrix_table = pd.DataFrame(
            {
                "true_positive": [tp_percentage],
                "false_positive": [fp_percentage],
                "true_negative": [tn_percentage],
                "false_negative": [fn_percentage],
            }
        )

        confusion_matrix_table_spark = spark.createDataFrame(confusion_matrix_table)
        confusion_matrix_table_spark.write.mode("overwrite").saveAsTable(
            confusion_matrix_table_path
        )
        LOGGER.info(
            "Confusion matrix written to table '%s' for run_id=%s",
            confusion_matrix_table_path,
            automl_run_id,
        )

    except mlflow.exceptions.MlflowException as e:
        raise RuntimeError(f"MLflow error while retrieving run {automl_run_id}: {e}")
    except Exception:
        LOGGER.exception("Failed to compute or store confusion matrix.")
        raise


def log_roc_table(
    institution_id: str,
    *,
    automl_run_id: str,
    catalog: str = "staging_sst_01",
    target_col: str = "target",
    modeling_dataset_name: str = "modeling_dataset",
    split_col: Optional[str] = None,
) -> None:
    """
    Computes and saves an ROC curve table (FPR, TPR, threshold, etc.) for a given MLflow run
    by reloading the test dataset and the trained model.

    Args:
        institution_id (str): Institution ID prefix for table name.
        automl_experiment_id (str): MLflow run ID of the trained model.
        experiment_id
        catalog (str): Destination catalog/schema for the ROC curve table.
    """
    try:
        from databricks.connect import DatabricksSession

        spark = DatabricksSession.builder.getOrCreate()
    except Exception:
        print("⚠️ Databricks Connect failed. Falling back to local Spark.")
        from pyspark.sql import SparkSession

        spark = (
            SparkSession.builder.master("local[*]").appName("Fallback").getOrCreate()
        )

    split_col = split_col or "split"

    table_path = f"{catalog}.{institution_id}_silver.training_{automl_run_id}_roc_curve"

    try:
        df = spark.read.table(modeling_dataset_name).toPandas()
        test_df = df[df[split_col] == "test"].copy()

        # Load model + features
        model = mlflow.sklearn.load_model(f"runs:/{automl_run_id}/model")
        feature_names: List[str] = model.named_steps["column_selector"].get_params()[
            "cols"
        ]

        # Prepare inputs for ROC
        y_true = test_df[target_col].values
        X_test = test_df[feature_names]
        y_scores = model.predict_proba(X_test)[:, 1]  # probabilities for class 1

        # Calculate ROC table manually and plot all thresholds.
        # Down the line, we might want to specify a threshold to reduce plot density
        thresholds = np.sort(np.unique(y_scores))[::-1]
        rounded_thresholds = sorted(
            set([round(t, 4) for t in thresholds]), reverse=True
        )

        P, N = np.sum(y_true == 1), np.sum(y_true == 0)

        rows = []
        for thresh in rounded_thresholds:
            y_pred = (y_scores >= thresh).astype(int)
            TP = np.sum((y_pred == 1) & (y_true == 1))
            FP = np.sum((y_pred == 1) & (y_true == 0))
            TN = np.sum((y_pred == 0) & (y_true == 0))
            FN = np.sum((y_pred == 0) & (y_true == 1))
            TPR = TP / P if P else 0
            FPR = FP / N if N else 0
            rows.append(
                {
                    "threshold": round(thresh, 4),
                    "true_positive_rate": round(TPR, 4),
                    "false_positive_rate": round(FPR, 4),
                    "true_positive": int(TP),
                    "false_positives": int(FP),
                    "true_negatives": int(TN),
                    "false_negatives": int(FN),
                }
            )

        roc_df = pd.DataFrame(rows)
        spark_df = spark.createDataFrame(roc_df)
        spark_df.write.mode("overwrite").saveAsTable(table_path)
        logging.info(
            "ROC table written to table '%s' for run_id=%s", table_path, automl_run_id
        )
    except Exception as e:
        raise RuntimeError(
            f"Failed to log ROC table for run {automl_run_id}: {e}"
        ) from e


############
## PLOTS! ##
############


def create_evaluation_plots(
    data: pd.DataFrame,
    risk_score_col: str,
    y_true_col: str,
    pos_label: PosLabelType,
    split_type: str,
) -> tuple[matplotlib.figure.Figure, ...]:
    """
    Create plots to evaluate a model overall - risk score histogram
    and calibration curve at low alert rates

    Args:
        data: containing predicted and actual outcome data
        risk_score_col: column name containing data of predicted risk scores
        y_true_col: column name containing data of actual outcome classes
        pos_label: label identifying the positive class in y_true
        split_type: type of data being plotted for labeling plots - train, test, or validation

    Returns:
        risk score histogram and calibration curve at low alert rates figures
    """
    title_suffix = f"{split_type} data - Overall"
    hist_fig = plot_support_score_histogram(data[risk_score_col], title_suffix)
    cal_fig = plot_calibration_curve(
        data[y_true_col], data[risk_score_col], "Overall", title_suffix, pos_label
    )
    return hist_fig, cal_fig


def create_evaluation_plots_by_subgroup(
    data: pd.DataFrame,
    risk_score_col: str,
    y_true_col: str,
    pos_label: PosLabelType,
    group_col: str,
    split_type: str,
) -> matplotlib.figure.Figure:
    """
    Create calibration curve plots to evaluate a model by group

    Args:
        data: containing predicted and actual outcome data, as well as group label
        risk_score_col: column name containing data of predicted risk scores
        y_true_col: column name containing data of actual outcome classes
        pos_label: label identifying the positive class in y_true
        group_col: column name containing data of subgroup labels
        split_type: type of data being plotted for labeling plots - train, test, or validation

    Returns:
        calibration curve figures by group
    """
    title_suffix = f"{split_type} data - {group_col}"

    grouped_data = data.groupby(group_col).agg(list)

    ys = grouped_data[y_true_col]
    scores = grouped_data[risk_score_col]

    subgroups = grouped_data.index
    n_counts = [len(subgroup_scores) for subgroup_scores in scores]
    names = [
        subgroup_name + f" (N={n})" for subgroup_name, n in zip(subgroups, n_counts)
    ]

    if len(subgroups) > 4:
        lowess_frac = 0.7
    else:
        lowess_frac = 0.6

    cal_subgroup_plot = plot_calibration_curve(
        ys, scores, names, title_suffix, pos_label, lowess_frac=lowess_frac
    )

    return cal_subgroup_plot


def plot_trained_models_comparison(
    automl_experiment_id: str, automl_metric: str
) -> matplotlib.figure.Figure:
    """
    Create a plot to evaluate all the models trained by AutoML.

    Args:
        automl_experiment_id: Experiment ID of the AutoML experiment
        automl_metric: Chosen AutoML optimization metric

    Returns:
        bar chart of model performance on test data by optimization metric.
    """
    df_sorted, metric_score_column = compare_trained_models(
        automl_experiment_id, automl_metric
    )

    df_sorted = df_sorted.sort_values(
        by=metric_score_column,
        ascending=True if not automl_metric == "log_loss" else False,
    )
    fig, ax = plt.subplots()

    ax.set_xlim(
        0,
        (
            df_sorted[metric_score_column].max() + 0.2
            if automl_metric == "log_loss"
            else 1
        ),
    )  # setting buffer after max log_loss to 0.2, for better viz

    colors = ["#A9A9A9"] * len(df_sorted)
    colors[-1] = "#5bc0de"  # Bright blue color for the best model

    bars = ax.barh(
        df_sorted["tags.model_type"], df_sorted[metric_score_column], color=colors
    )

    for bar in bars:
        ax.text(
            bar.get_width()
            - (
                max(df_sorted[metric_score_column]) * 0.005
            ),  # Position the text inside the bar
            bar.get_y() + bar.get_height() / 2,
            f"{bar.get_width():.4f}",
            va="center",
            ha="right",
            fontweight="bold",
            color="white",
            fontsize=12,
        )

    sort_order = (
        "Lowest to Highest" if automl_metric == "log_loss" else "Highest to Lowest"
    )
    automl_metric = (
        automl_metric
        if automl_metric == "log_loss"
        else f"{automl_metric.capitalize()} Score"
    )

    ax.set(
        title=f"{automl_metric} by Model Type {sort_order}",
        facecolor="none",
        frame_on=False,
    )
    ax.tick_params(axis="y", which="both", left=False, right=False, labelleft=True)
    ax.tick_params(axis="x", colors="lightgrey", which="both")  # Color of ticks
    ax.xaxis.grid(True, color="lightgrey", linestyle="--", linewidth=0.5)
    fig.tight_layout()

    mlflow.log_figure(fig, "model_comparison.png")
    plt.close()

    return fig


def plot_calibration_curve(
    y_true: pd.Series,
    risk_score: pd.Series,
    keys: str | list[str],
    title_suffix: str,
    pos_label: PosLabelType,
    lowess_frac: t.Optional[float] = None,
) -> matplotlib.figure.Figure:
    """
    Plot calibration curve.

    Args:
        y_true (array-like of shape (n_samples,) or (n_groups,)): overall or group-level true outcome class
        risk_score (array-like of shape (n_samples,) or (n_groups,)): overall or group level predicted risk scores
        keys: overall or subgroup level labels for labeling lines
        title_suffix: suffix for plot title
        pos_label: label identifying the positive class. Defaults to True.

    Returns:
        line plot of prediction bins X fraction of positive class
    """
    if not _check_array_of_arrays(y_true):
        y_true = [y_true]
        risk_score = [risk_score]
        keys = [keys]  # type: ignore

    fig, ax = plt.subplots()

    for j in range(len(y_true)):
        prob_true, prob_pred = calibration_curve(
            y_true[j],
            risk_score[j],
            n_bins=10,
            strategy="uniform",
            pos_label=pos_label,  # type: ignore
        )
        if lowess_frac:
            # When we create calibration curves with less data points (i.e. for subgroups),
            # it can look choppy and be difficult to interpret. We use locally weighted
            # scatterplot smoothing (LOWESS) to fit a smooth curve to the data points. LOWESS
            # is a non-parametric smoother that tries to find a curve of best fit without
            # assuming the data fits a particular shape.
            # Resource: https://www.statisticshowto.com/lowess-smoothing/
            prob_true = lowess(
                endog=prob_true,
                exog=prob_pred,
                frac=lowess_frac,
                it=0,
                is_sorted=False,
                return_sorted=False,
            )
        sns.lineplot(
            x=prob_pred, y=prob_true, color=PALETTE[j + 1], ax=ax, label=keys[j]
        )
    sns.lineplot(
        x=[0, 1],
        y=[0, 1],
        linestyle="dashed",
        color=PALETTE[0],
        ax=ax,
        label="Perfectly calibrated",
    )

    ax.set(
        xlabel="Mean predicted value",
        ylabel="Fraction of positives",
        title=f"Calibration Curve - {title_suffix}",
    )
    ax.set_xlim(left=-0.05, right=1.05)
    ax.set_ylim(bottom=-0.05, top=1.05)
    ax.legend(loc="lower right")
    return fig


def plot_support_score_histogram(
    support_scores: str | Sequence, title_suffix: str
) -> matplotlib.figure.Figure:
    """
    Plot histogram of support scores.

    Args:
        support_scores: support scores
        title_suffix: suffix for plot title
    """
    fig, ax = plt.subplots()
    sns.histplot(x=support_scores, ax=ax, color=PALETTE[1])
    ax.set(
        xlabel="Support Score", title=f"Distribution of support scores - {title_suffix}"
    )
    return fig


def plot_features_permutation_importance(
    result: sklearn.utils.Bunch, feature_cols: pd.Index
) -> matplotlib.figure.Figure:
    """
    Plot features' permutation importances (computed previously).

    Args:
        result
        feature_cols

    See Also:
        - :func:`compute_feature_permutation_importance()`
    """
    sorted_importances_idx = result.importances_mean.argsort()
    df_importances = pd.DataFrame(
        data=result.importances[sorted_importances_idx].T,
        columns=feature_cols[sorted_importances_idx],
    )

    fig, ax = plt.subplots(layout="constrained", figsize=(10, 10))
    df_importances.plot.box(vert=False, whis=10, ax=ax)
    ax.set_title("Permutation Feature Importances")
    ax.axvline(x=0, color="k", linestyle="--")
    ax.set_xlabel("decrease in model score")
    return fig


def plot_shap_beeswarm(shap_values):
    # Define the color palette
    start_color = (0.34, 0.79, 0.55)  # Green
    middle_color = (0.98, 0.82, 0.22)  # Yellow
    end_color = (0.95, 0.60, 0.19)  # Orange
    colors = [start_color, middle_color, end_color]
    cmap = mcolors.LinearSegmentedColormap.from_list("custom_cmap", colors, 256)

    plt.figure(figsize=(5, 6))  # Adjust height as needed.

    # Plot the beeswarm
    ax = shap.plots.beeswarm(
        shap_values,
        show=False,
        max_display=25,
        color=cmap,
        color_bar=False,
        plot_size=(5, 25),
    )

    # Get the current y-axis tick labels (feature names)
    y_ticks_labels = [label.get_text() for label in ax.get_yticklabels()]

    # Modify the feature names
    modified_labels = [label.replace("_", " ").title() for label in y_ticks_labels]

    # Set the modified labels back to the y-axis
    ax.set_yticklabels(modified_labels)

    # Add horizontal lines
    # Get the y-axis tick positions
    y_ticks = ax.get_yticks()
    for y in y_ticks:
        # Add a dashed gray line
        ax.axhline(y=y, color="gray", linestyle="-", linewidth=0.5)

    # Add colorbar above chart
    cbar = plt.colorbar(
        plt.cm.ScalarMappable(
            cmap=cmap,
        ),
        ax=ax,
        location="top",
        shrink=0.5,
        aspect=20,
        ticks=[],
        pad=0.02,
    )
    cbar.outline.set_visible(False)
    cbar.set_label("Feature Value", loc="center", labelpad=10)
    # Add labels to the left and right of the colorbar
    cbar.ax.text(
        -0.1,  # Adjust horizontal position for left label
        0.5,  # Vertical position (middle of colorbar)
        "Low",  # Left label text
        va="center",
        ha="right",
        transform=cbar.ax.transAxes,
    )
    cbar.ax.text(
        1.1,  # Adjust horizontal position for right label
        0.5,  # Vertical position (middle of colorbar)
        "High",  # Right label text
        va="center",
        ha="left",
        transform=cbar.ax.transAxes,
    )
    return plt.gcf()


def _check_array_of_arrays(input_array: pd.Series) -> bool:
    """
    Check if an input array contains sub-arrays, and return True if so; False otherwise.
    Used for plotting different groups of predictions.
    """
    try:
        assert isinstance(input_array, pd.Series)
        assert isinstance(input_array[0], list)
        return True
    except Exception:
        return False



import collections

import numpy as np
import pandas as pd
import pytest
from pandas.api.types import is_numeric_dtype
from unittest.mock import patch

from student_success_tool.modeling.inference import (
    _get_mapped_feature_name,
    calculate_shap_values,
    calculate_shap_values_spark_udf,
    select_top_features_for_display,
    generate_ranked_feature_table,
    top_shap_features,
    support_score_distribution_table,
)


class DummyKernelExplainer:
    def __call__(self, X):
        Explanation = collections.namedtuple("Explanation", ["values", "feature_names"])
        return Explanation(np.random.rand(len(X), len(X.columns)) * 0.1, X.columns)

    def shap_values(self, X):
        # for simplicity, return random numbers of the proper shape
        return np.random.rand(len(X), len(X.columns)) * 0.1


@pytest.fixture(scope="module")
def explainer():
    return DummyKernelExplainer()


@pytest.mark.parametrize(
    [
        "features",
        "unique_ids",
        "predicted_probabilities",
        "shap_values",
        "n_features",
        "needs_support_threshold_prob",
        "features_table",
        "exp",
    ],
    [
        (
            pd.DataFrame(
                {
                    "x1": ["val1", "val2", "val3"],
                    "x2": [True, False, True],
                    "x3": [2.0, 1.0001, 0.5],
                    "x4": [1, 2, 3],
                }
            ),
            pd.Series([1, 2, 3]),
            [0.9, 0.1, 0.5],
            np.array(
                [[1.0, 0.9, 0.8, 0.7], [0.0, -1.0, 0.9, -0.8], [0.25, 0.0, -0.5, 0.75]]
            ),
            3,
            0.5,
            {
                "x1": {"name": "feature #1"},
                "x2": {"name": "feature #2"},
                "x3": {"name": "feature #3"},
            },
            pd.DataFrame(
                {
                    "Student ID": [1, 2, 3],
                    "Support Score": [0.9, 0.1, 0.5],
                    "Support Needed": [True, False, True],
                    "Feature_1_Name": ["feature #1", "feature #2", "x4"],
                    "Feature_1_Value": ["val1", "False", "3"],
                    "Feature_1_Importance": [1.0, -1.0, 0.75],
                    "Feature_2_Name": ["feature #2", "feature #3", "feature #3"],
                    "Feature_2_Value": ["True", "1.0", "0.5"],
                    "Feature_2_Importance": [0.9, 0.9, -0.5],
                    "Feature_3_Name": ["feature #3", "x4", "feature #1"],
                    "Feature_3_Value": ["2.0", "2", "val3"],
                    "Feature_3_Importance": [0.8, -0.8, 0.25],
                }
            ),
        ),
        (
            pd.DataFrame(
                {
                    "x1": ["val1", "val2", "val3"],
                    "x2": [True, False, True],
                    "x3": [2.0, 1.0, 0.5],
                    "x4": [1, 2, 3],
                }
            ),
            pd.Series([1, 2, 3]),
            [0.9, 0.1, 0.5],
            np.array(
                [[1.0, 0.9, 0.8, 0.7], [0.0, -1.0, 0.9, -0.8], [0.25, 0.0, -0.5, 0.75]]
            ),
            1,
            None,
            None,
            pd.DataFrame(
                {
                    "Student ID": [1, 2, 3],
                    "Support Score": [0.9, 0.1, 0.5],
                    "Feature_1_Name": ["x1", "x2", "x4"],
                    "Feature_1_Value": ["val1", "False", "3"],
                    "Feature_1_Importance": [1.0, -1.0, 0.75],
                }
            ),
        ),
    ],
)
def test_select_top_features_for_display(
    features,
    unique_ids,
    predicted_probabilities,
    shap_values,
    n_features,
    needs_support_threshold_prob,
    features_table,
    exp,
):
    obs = select_top_features_for_display(
        features,
        unique_ids,
        predicted_probabilities,
        shap_values,
        n_features=n_features,
        needs_support_threshold_prob=needs_support_threshold_prob,
        features_table=features_table,
    )
    assert isinstance(obs, pd.DataFrame) and not obs.empty
    assert pd.testing.assert_frame_equal(obs, exp) is None


@pytest.mark.parametrize(
    ["df", "feature_names", "fillna_values", "student_id_col", "exp_shape"],
    [
        (
            pd.DataFrame(
                {
                    "student_id": [1, 2, 3],
                    "feature1": [0.1, 0.2, 0.3],
                    "feature2": [0.4, 0.5, 0.6],
                }
            ),
            ["feature1", "feature2"],
            pd.Series([0.2, 0.5]),
            "student_id",
            (3, 3),
        ),
    ],
)
def test_calculate_shap_values(
    explainer, df, feature_names, fillna_values, student_id_col, exp_shape
):
    obs = calculate_shap_values(
        df,
        explainer,
        feature_names=feature_names,
        fillna_values=fillna_values,
        student_id_col=student_id_col,
    )
    assert isinstance(obs, pd.DataFrame) and not obs.empty
    assert obs.shape == exp_shape
    assert student_id_col in obs.columns
    assert all(is_numeric_dtype(obs[feature_name]) for feature_name in feature_names)
    assert obs[student_id_col].equals(df[student_id_col])


@pytest.mark.parametrize(
    "input_data, expected_shape",
    [
        (
            {
                "student_id": [1, 2, 3],
                "feature1": [0.1, 0.2, 0.3],
                "feature2": [0.4, 0.5, 0.6],
            },
            (3, 3),
        ),
        (
            {"student_id": [1, 2], "feature1": [0.1, 0.2], "feature2": [0.4, 0.5]},
            (2, 3),
        ),
    ],
)
def test_calculate_shap_values_spark_udf_basic(input_data, expected_shape, explainer):
    df = pd.DataFrame(input_data)
    student_id_col = "student_id"
    model_features = ["feature1", "feature2"]
    mode = df.mode().iloc[0]

    iterator = iter([df])

    result = list(
        calculate_shap_values_spark_udf(
            iterator,
            student_id_col=student_id_col,
            model_features=model_features,
            explainer=explainer,
            mode=mode,
        )
    )

    # Check that the result contains the expected number of rows and columns
    shap_df = result[0]
    assert shap_df.shape == expected_shape

    # Ensure that 'student_id' column is present
    assert student_id_col in shap_df.columns

    # Ensure that SHAP values are generated and are numeric
    assert is_numeric_dtype(shap_df[model_features].iloc[0, 0])
    assert is_numeric_dtype(shap_df[model_features].iloc[0, 1])

    # Ensure student IDs are correctly reattached
    assert shap_df[student_id_col].iloc[0] == 1
    assert shap_df[student_id_col].iloc[1] == 2


@pytest.mark.parametrize(
    "batch1_data, batch2_data, expected_shape1, expected_shape2",
    [
        (
            {
                "student_id": [1, 2, 3],
                "feature1": [0.1, 0.2, 0.3],
                "feature2": [0.4, 0.5, 0.6],
            },
            {
                "student_id": [4, 5, 6],
                "feature1": [0.7, 0.8, 0.9],
                "feature2": [0.6, 0.7, 0.8],
            },
            (3, 3),
            (3, 3),
        ),
        (
            {
                "student_id": [4, 5, 6],
                "feature1": [0.1, 0.2, 0.3],
                "feature2": [0.4, 0.5, 0.6],
            },
            {
                "student_id": [4, 5, 6],
                "feature1": [0.5, 0.6, 0.7],
                "feature2": [0.7, 0.8, 0.9],
            },
            (3, 3),
            (3, 3),
        ),
    ],
)
def test_calculate_shap_values_spark_udf_multiple_batches(
    batch1_data, batch2_data, expected_shape1, expected_shape2, explainer
):
    batch1 = pd.DataFrame(batch1_data)
    batch2 = pd.DataFrame(batch2_data)

    student_id_col = "student_id"
    model_features = ["feature1", "feature2"]
    mode = batch1.mode().iloc[0]

    iterator = iter([batch1, batch2])

    result = list(
        calculate_shap_values_spark_udf(
            iterator,
            student_id_col=student_id_col,
            model_features=model_features,
            explainer=explainer,
            mode=mode,
        )
    )

    # Ensure we have two DataFrames
    assert len(result) == 2

    # Check first batch
    shap_df1 = result[0]
    assert shap_df1.shape == expected_shape1

    # Check second batch
    shap_df2 = result[1]
    assert shap_df2.shape == expected_shape2


@pytest.fixture
def ranked_feature_table_data():
    features = pd.DataFrame(
        {
            "pell_status": [False, True, False],
            "english_math_gateway": ["E", "M", "M"],
            "term_gpa": [4.0, 3.0, 2.0],
        }
    )
    shap_values = np.array(
        [
            [0.1, -0.2, 0.05],
            [0.3, 0.1, -0.05],
            [0.2, -0.1, 0.15],
        ]
    )
    features_table = {
        "pell_status": {"name": "Pell Status"},
        "english_math_gateway": {"name": "English or Math Gateway"},
        "term_gpa": {"name": "Term GPA"},
    }
    return features, shap_values, features_table


@pytest.mark.parametrize("use_features_table", [True, False])
def test_generate_ranked_feature_table(ranked_feature_table_data, use_features_table):
    features, shap_values, features_table = ranked_feature_table_data

    selected_features_table = features_table if use_features_table else None

    result = generate_ranked_feature_table(
        features, shap_values, selected_features_table
    )

    assert isinstance(result, pd.DataFrame) and not result.empty
    assert set(result.columns) == {
        "Feature Name",
        "Data Type",
        "Average SHAP Magnitude",
    }

    # Verify descending sort order by Average SHAP Magnitude
    assert result["Average SHAP Magnitude"].is_monotonic_decreasing

    if use_features_table:
        assert "English or Math Gateway" in result["Feature Name"].values
    else:
        assert "term_gpa" in result["Feature Name"].values


@pytest.mark.parametrize(
    ["feature_col", "features_table", "exp"],
    [
        (
            "academic_term",
            {"academic_term": {"name": "academic term"}},
            "academic term",
        ),
        ("foo_bar", {"academic_term": {"name": "academic term"}}, "foo_bar"),
        (
            "num_courses_course_subject_area_24",
            {
                r"num_courses_course_subject_area_(\d+)": {
                    "name": "number of courses taken in subject area {} this term"
                }
            },
            "number of courses taken in subject area 24 this term",
        ),
        (
            "num_courses_course_id_engl_101",
            {
                r"num_courses_course_id_(.*)": {
                    "name": "number of times course '{}' taken this term"
                }
            },
            "number of times course 'engl_101' taken this term",
        ),
        (
            "num_courses_course_id_engl_101_cumfrac",
            {
                r"num_courses_course_id_(.*)_cumfrac": {
                    "name": "fraction of times course '{}' taken so far"
                }
            },
            "fraction of times course 'engl_101' taken so far",
        ),
    ],
)
def test_get_mapped_feature_name(feature_col, features_table, exp):
    obs = _get_mapped_feature_name(feature_col, features_table)
    assert isinstance(obs, str)
    assert obs == exp


@pytest.fixture
def sample_data():
    features = pd.DataFrame(
        {
            "feature1": [1, 2, 3],
            "feature2": [4, 5, 6],
            "feature3": [7, 8, 9],
            "feature4": [0, 1, 0],
            "feature5": [2, 2, 2],
            "feature6": [1, 1, 1],
            "feature7": [0, 0, 1],
            "feature8": [3, 3, 3],
            "feature9": [4, 4, 4],
            "feature10": [5, 5, 5],
            "feature11": [6, 6, 6],
        }
    )
    unique_ids = pd.Series([101, 102, 103])
    shap_values = np.array(
        [
            [0.1, 0.3, 0.2, 0.0, 0.4, 0.1, 0.0, 0.3, 0.2, 0.5, 0.6],
            [0.2, 0.2, 0.1, 0.0, 0.3, 0.0, 0.1, 0.2, 0.3, 0.4, 0.5],
            [0.3, 0.1, 0.3, 0.0, 0.2, 0.0, 0.0, 0.1, 0.4, 0.3, 0.4],
        ]
    )
    features_table = {
        "feature1": {
            "name": "Feature 1 Name",
            "short_desc": "A short description of feature 1",
            "long_desc": "A long description of feature 1",
        },
        "feature2": {
            "name": "Feature 2 Name",
            "short_desc": "A short description of feature 2",
            "long_desc": "A long description of feature 2",
        },
        "feature3": {
            "name": "Feature 3 Name",
        },
    }
    return features, unique_ids, shap_values, features_table


def test_top_shap_features_behavior(sample_data):
    features, unique_ids, shap_values, features_table = sample_data
    result = top_shap_features(
        features, unique_ids, shap_values, features_table=features_table
    )

    # Check output shape and columns
    assert isinstance(result, pd.DataFrame)
    assert set(result.columns) == {
        "student_id",
        "feature_name",
        "shap_value",
        "feature_value",
        "feature_readable_name",
        "feature_short_desc",
        "feature_long_desc",
    }

    # Check top 10 feature selection
    top_features = result["feature_name"].unique()
    assert len(top_features) == 10

    grouped = result.groupby("feature_readable_name")["shap_value"].apply(
        lambda x: np.mean(np.abs(x))
    )
    shap_values = grouped.sort_values(ascending=False).values
    assert all(
        shap_values[i] >= shap_values[i + 1] for i in range(len(shap_values) - 1)
    )

    assert len(grouped) == 10
    print(grouped)
    assert grouped.index[0] == "Feature 1 Name"
    assert grouped.index[1] == "Feature 2 Name"

    assert (
        result["feature_short_desc"]
        .apply(lambda x: isinstance(x, str) or x is None)
        .all()
    )
    assert (
        result["feature_long_desc"]
        .apply(lambda x: isinstance(x, str) or x is None)
        .all()
    )


def test_handles_fewer_than_10_features():
    features = pd.DataFrame(
        {
            "feature1": [1, 2],
            "feature2": [3, 4],
        }
    )
    unique_ids = pd.Series([1, 2])
    shap_values = np.array([[0.5, 0.1], [0.3, 0.4]])

    result = top_shap_features(features, unique_ids, shap_values)
    assert set(result["feature_name"].unique()) == {"feature1", "feature2"}
    assert len(result) == 4  # 2 students × 2 features


def test_empty_input():
    features = pd.DataFrame()
    unique_ids = pd.Series(dtype=int)
    shap_values = np.empty((0, 0))

    with pytest.raises(ValueError):
        top_shap_features(features, unique_ids, shap_values)


@patch("student_success_tool.modeling.inference.select_top_features_for_display")
@pytest.mark.parametrize(
    [
        "features",
        "unique_ids",
        "predicted_probabilities",
        "shap_values",
        "n_features",
        "needs_support_threshold_prob",
        "features_table",
        "exp",
    ],
    [
        (
            pd.DataFrame(
                {
                    "x1": ["val1", "val2", "val3"],
                    "x2": [True, False, True],
                    "x3": [2.0, 1.0001, 0.5],
                    "x4": [1, 2, 3],
                }
            ),
            pd.Series([1, 2, 3]),
            [0.9, 0.1, 0.5],
            np.array(
                [
                    [1.0, 0.9, 0.8, 0.7],
                    [0.0, -1.0, 0.9, -0.8],
                    [0.25, 0.0, -0.5, 0.75],
                ]
            ),
            3,
            0.5,
            {
                "x1": {"name": "feature #1"},
                "x2": {"name": "feature #2"},
                "x3": {"name": "feature #3"},
            },
            pd.DataFrame(
                {
                    "Student ID": [1, 2, 3],
                    "Support Score": [0.9, 0.1, 0.5],
                    "Support Needed": [True, False, True],
                    "Feature_1_Name": ["feature #1", "feature #2", "x4"],
                    "Feature_1_Value": ["val1", "False", "3"],
                    "Feature_1_Importance": [1.0, -1.0, 0.75],
                    "Feature_2_Name": ["feature #2", "feature #3", "feature #3"],
                    "Feature_2_Value": ["True", "1.0", "0.5"],
                    "Feature_2_Importance": [0.9, 0.9, -0.5],
                    "Feature_3_Name": ["feature #3", "x4", "feature #1"],
                    "Feature_3_Value": ["2.0", "2", "val3"],
                    "Feature_3_Importance": [0.8, -0.8, 0.25],
                }
            ),
        )
    ],
)
def test_support_score_distribution_table(
    mock_select_top_features_for_display,
    features,
    unique_ids,
    predicted_probabilities,
    shap_values,
    n_features,
    needs_support_threshold_prob,
    features_table,
    exp,
):
    inference_params = {
        "num_top_features": n_features,
        "min_prob_pos_label": needs_support_threshold_prob or 0.0,
    }

    mock_select_top_features_for_display.return_value = exp

    result = support_score_distribution_table(
        df_serving=features,
        unique_ids=unique_ids,
        pred_probs=predicted_probabilities,
        shap_values=pd.DataFrame(shap_values),
        inference_params=inference_params,
        features_table=features_table,
    )

    assert isinstance(result, pd.DataFrame)
    assert set(result.columns) == {
        "bin_lower",
        "bin_upper",
        "support_score",
        "count_of_students",
        "pct",
    }
    assert result["count_of_students"].sum() == len(unique_ids)
    assert np.isclose(result["pct"].sum(), 100.0, atol=0.01)

    # Binning logic checks
    for _, row in result.iterrows():
        expected_midpoint = round((row["bin_lower"] + row["bin_upper"]) / 2, 2)
        assert row["support_score"] == expected_midpoint
        assert round(row["bin_upper"] - row["bin_lower"], 2) == 0.1
        assert 0.1 <= row["bin_lower"] < row["bin_upper"] <= 1.0



import mlflow.tracking
import pandas as pd
import pytest

from student_success_tool.modeling import evaluation


@pytest.fixture
def patch_mlflow(monkeypatch):
    def _patch(mock_df, target="mlflow.search_runs"):
        monkeypatch.setattr(target, lambda *args, **kwargs: mock_df)

    return _patch


def test_check_array_of_arrays_true():
    input_array = pd.Series([[1, 0, 1], [0, 1, 0]])
    assert evaluation._check_array_of_arrays(input_array)


def test_check_array_of_arrays_false():
    input_array = pd.Series([1, 0, 1])
    assert not evaluation._check_array_of_arrays(input_array)


@pytest.mark.parametrize(
    ["y_true", "risk_score", "q", "sensitivity"],
    [
        (["Y", "Y", "Y", "N", "N"], [0.12, 0.32, 0.98, 0.48, 0.87], 0.99, 1 / 3),
        (
            ["N", "N", "N", "N", "Y", "N", "N", "N", "N", "Y"],
            [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1],
            0.9,
            0.5,
        ),
    ],
)
def test_get_sensitivity_of_top_q_pctl_thresh(y_true, risk_score, q, sensitivity):
    assert (
        evaluation.get_sensitivity_of_top_q_pctl_thresh(
            y_true, risk_score, q, pos_label="Y"
        )
        == sensitivity
    )


@pytest.mark.parametrize(
    ["data", "metric", "expected_order", "expected_columns"],
    [
        (
            {
                "tags.model_type": ["Model A", "Model B", "Model C"],
                "metrics.val_recall_score": [0.92, 0.88, 0.93],
            },
            "recall",
            ["Model C", "Model A", "Model B"],
            ["tags.model_type", "metrics.val_recall_score"],
        ),
        (
            {
                "tags.model_type": ["Model D", "Model E", "Model F"],
                "metrics.val_f1_score": [0.80, 0.78, 0.81],
            },
            "f1",
            ["Model F", "Model D", "Model E"],
            ["tags.model_type", "metrics.val_f1_score"],
        ),
        (
            {
                "tags.model_type": ["Model G", "Model H", "Model I"],
                "metrics.val_log_loss": [0.4521, 0.3501, 0.5502],
            },
            "log_loss",
            ["Model H", "Model G", "Model I"],
            ["tags.model_type", "metrics.val_log_loss"],
        ),
    ],
)
def test_compare_trained_models(
    data, metric, expected_order, expected_columns, patch_mlflow, monkeypatch
):
    def _search_runs_patch(experiment_ids, output_format):
        return pd.DataFrame(data)

    monkeypatch.setattr(mlflow, "search_runs", _search_runs_patch)
    result, _ = evaluation.compare_trained_models("dummy_id", metric)
    print(result["tags.model_type"].tolist())
    assert isinstance(result, pd.DataFrame), "The result should be a pandas DataFrame."
    assert result["tags.model_type"].tolist() == expected_order, (
        "Models are not sorted in ascending order based on the metric."
    )
    assert all(column in result.columns for column in expected_columns), (
        "DataFrame should contain specific columns."
    )


@pytest.fixture
def mock_runs_df():
    """Mock MLflow run data for testing."""
    return pd.DataFrame(
        {
            "run_id": ["r1", "r2", "r3"],
            "tags.mlflow.runName": ["run_1", "run_2", "run_3"],
            "metrics.test_roc_auc": [0.80, 0.60, 0.90],
            "metrics.test_recall_score": [0.70, 0.95, 0.60],
            "metrics.val_log_loss": [0.25, 0.20, 0.30],
        }
    )


@pytest.mark.parametrize(
    "metrics, expected_run_name",
    [
        (["test_roc_auc"], "run_3"),
        (["test_recall_score"], "run_2"),
        (["val_log_loss"], "run_2"),
        (["test_roc_auc", "val_log_loss"], "run_1"),
    ],
)
def test_get_top_runs_balanced(metrics, expected_run_name, mock_runs_df, patch_mlflow):
    mock_df = pd.DataFrame(
        {
            "run_id": ["r1", "r2", "r3"],
            "tags.mlflow.runName": ["run_1", "run_2", "run_3"],
            "metrics.test_roc_auc": [0.80, 0.60, 0.90],
            "metrics.test_recall_score": [0.70, 0.95, 0.60],
            "metrics.val_log_loss": [0.25, 0.20, 0.30],
        }
    )

    patch_mlflow(mock_df)

    top = evaluation.get_top_runs(
        experiment_id="dummy",
        optimization_metrics=metrics,
        topn_runs_included=1,
    )

    assert list(top.keys())[0] == expected_run_name


@pytest.mark.parametrize(
    "metrics, expected_top",
    [
        (["test_recall_score"], "run_2"),
        (["val_log_loss"], "run_2"),
        (["test_roc_auc"], "run_3"),
        (["test_roc_auc", "val_log_loss"], "run_1"),
    ],
)
def test_get_top_runs_parametrized(metrics, expected_top, mock_runs_df, patch_mlflow):
    patch_mlflow(mock_runs_df)

    top = evaluation.get_top_runs(
        experiment_id="dummy",
        optimization_metrics=metrics,
        topn_runs_included=1,
    )

    assert list(top.keys())[0] == expected_top



import pandas as pd
import pytest

from student_success_tool.preprocessing.selection import pdp


@pytest.fixture(scope="module")
def df_test():
    return pd.DataFrame(
        {
            "student_id": ["01", "01", "01", "02", "02", "03", "04", "05"],
            "cohort_id": [
                "2020-21 SPRING",
                "2020-21 SPRING",
                "2020-21 SPRING",
                "2021-22 FALL",
                "2021-22 FALL",
                "2019-20 FALL",
                "2020-21 SPRING",
                "2022-23 FALL",
            ],
            "term_id": [
                "2020-21 FALL",
                "2020-21 SPRING",
                "2021-22 FALL",
                "2021-22 FALL",
                "2021-22 SPRING",
                "2019-20 SPRING",
                "2020-21 SPRING",
                "2022-23 FALL",
            ],
            "credential_sought": [
                "Associate's",
                "Associate's",
                "Associate's",
                "Bachelor's",
                "Bachelor's",
                "Associate's",
                "Associate's",
                pd.NA,
            ],
            "enrollment_type": [
                "FIRST-TIME",
                "FIRST-TIME",
                "FIRST-TIME",
                "FIRST-TIME",
                "FIRST-TIME",
                "FIRST-TIME",
                "TRANSFER-IN",
                pd.NA,
            ],
            "enrollment_intensity": [
                "FULL-TIME",
                "FULL-TIME",
                "FULL-TIME",
                "PART-TIME",
                "PART-TIME",
                "PART-TIME",
                "FULL-TIME",
                pd.NA,
            ],
            "num_credits_earned": [25.0, 30.0, 35.0, 25.0, 35.0, 20.0, 45.0, 10.0],
            "term_rank": [3, 4, 5, 5, 6, 2, 4, 8],
            "term_is_pre_cohort": [
                True,
                False,
                False,
                False,
                False,
                False,
                False,
                False,
            ],
        },
    ).astype(
        {
            "student_id": "string",
            "credential_sought": "string",
            "enrollment_type": "string",
            "enrollment_intensity": "string",
        }
    )


@pytest.mark.parametrize(
    ["criteria", "exp"],
    [
        (
            {
                "credential_sought": "Associate's",
                "enrollment_type": "FIRST-TIME",
                "enrollment_intensity": {"FULL-TIME", "PART-TIME"},
            },
            pd.DataFrame(
                data={
                    "credential_sought": ["Associate's", "Associate's"],
                    "enrollment_type": ["FIRST-TIME", "FIRST-TIME"],
                    "enrollment_intensity": ["FULL-TIME", "PART-TIME"],
                },
                index=pd.Index(["01", "03"], name="student_id", dtype="string"),
            ).astype(
                {
                    "credential_sought": "string",
                    "enrollment_type": "string",
                    "enrollment_intensity": "string",
                }
            ),
        ),
        (
            {
                "enrollment_type": "FIRST-TIME",
                "enrollment_intensity": "PART-TIME",
            },
            pd.DataFrame(
                data={
                    "enrollment_type": ["FIRST-TIME", "FIRST-TIME"],
                    "enrollment_intensity": ["PART-TIME", "PART-TIME"],
                },
                index=pd.Index(["02", "03"], name="student_id", dtype="string"),
            ).astype({"enrollment_type": "string", "enrollment_intensity": "string"}),
        ),
        (
            {"credential_sought": ["Associate's", "Bachelor's"]},
            pd.DataFrame(
                data={
                    "credential_sought": [
                        "Associate's",
                        "Bachelor's",
                        "Associate's",
                        "Associate's",
                    ],
                },
                index=pd.Index(
                    ["01", "02", "03", "04"], name="student_id", dtype="string"
                ),
            ).astype({"credential_sought": "string"}),
        ),
        (
            {"enrollment_type": "RE-ADMIT"},
            pd.DataFrame(
                data={"enrollment_type": []},
                index=pd.Index([], name="student_id", dtype="string"),
            ).astype("string"),
        ),
    ],
)
def test_select_students_by_attributes(df_test, criteria, exp):
    obs = pdp.select_students_by_attributes(
        df_test, student_id_cols="student_id", **criteria
    )
    assert isinstance(obs, pd.DataFrame)
    assert pd.testing.assert_frame_equal(obs, exp) is None



"""
Utility functions that interact with Google Cloud Storage.
For instance, copying files from Databricks volumes into GCP buckets.
"""

import os

from databricks.sdk.runtime import dbutils  # noqa: F401
from google.cloud import storage
from google.cloud.storage.bucket import Bucket


def save_file(
    bucket: Bucket, src_volume_filepath: str, dest_bucket_pathname: str
) -> None:
    """Save file from databricks volume to a GCP bucket path.

    Args:
      bucket: The bucket object.
      src_volume_filepath: The source filepath to the Databricks volume.
      dest_bucket_pathname: The destination filepath in GCP.

    Returns:
      Nothing.
    """
    blob = bucket.blob(dest_bucket_pathname)
    if blob.exists():
        raise ValueError(dest_bucket_pathname + ": File already exists in bucket.")
    blob.upload_from_filename(src_volume_filepath)


def publish_inference_output_files(
    db_workspace: str,
    institution_name: str,
    external_bucket_name: str,
    sst_job_id: str,
    approved: bool,
) -> str:
    """Publish output files to bucket, with folder determined based on approved parameter,
    and return the status of the job as a string.

    Args:
      db_workspace: The Databricks workspace to get files from.
      institution_name: The Databricks institution name substring used in the schema (e.g. it would be
        'uni_of_datakind' if the gold schema was 'uni_of_datakind_gold') -- this should match the
        "Databricksified" name generated by the webapp.
      external_bucket_name: The destination bucket in GCP.
      sst_job_id: the job run id of this task.
      approved: whether this file should be published in an approved state or not.

    Returns:
      The status string of the job run from Databricks.
    """
    # Destination: The output directory in the bucket that the files get moved to.
    bucket_directory = f"{'approved' if approved else 'unapproved'}/{sst_job_id}"
    # Source: The Databricks volume path we want to read the files from.
    volume_path_top_level = f"/Volumes/{db_workspace}/{institution_name}_gold/gold_volume/inference_jobs/{sst_job_id}/ext"
    volume_path_inference_folder = f"{volume_path_top_level}/inference_output"

    storage_client = storage.Client()
    bucket = storage_client.bucket(external_bucket_name)

    # Full filepaths for the Databricks volume.
    files_to_move = []
    status_string = ""
    # There will be other files in the inference_output folder generated by Databricks
    # including time started/completed, and completion status.
    for f in dbutils.fs.ls(volume_path_inference_folder):
        filename = f.name
        if filename.startswith("_"):
            if not filename.startswith("_committed") and not filename.startswith(
                "_started"
            ):
                status_string = filename
        elif filename.endswith(".csv"):
            files_to_move.append(f"{volume_path_inference_folder}/{filename}")
    # The top level volume directory can include other files like pngs etc. that are part
    # of the published set of files. So we still need to iterate over these files.
    for f in dbutils.fs.ls(volume_path_top_level):
        if f.name.endswith(".png"):
            files_to_move.append(f"{volume_path_top_level}/{f.name}")

    for f in files_to_move:
        new_fname = os.path.basename(f)
        if new_fname.endswith(".csv"):
            new_fname = "inference_output.csv"
        save_file(
            bucket,
            f,
            f"{bucket_directory}/{new_fname}",
        )
    return status_string



import pandas as pd
import pytest

from student_success_tool.preprocessing.features.pdp import constants, course


@pytest.mark.parametrize(
    ["df", "min_passing_grade", "course_level_pattern", "exp"],
    [
        (
            pd.DataFrame(
                {
                    "academic_year": ["2020-21", "2020-21", "2021-22"],
                    "academic_term": ["FALL", "SPRING", "FALL"],
                    "course_prefix": ["MATH", "MATH", "PHYS"],
                    "course_number": ["101", "202", "303"],
                    "course_cip": ["40.02", "45.06", "03.01"],
                    "course_type": ["CU", "CU", "CC"],
                    "delivery_method": ["O", "H", "F"],
                    "grade": ["4", "1", "W"],
                }
            ).astype({"grade": "string"}),
            2.0,
            constants.DEFAULT_COURSE_LEVEL_PATTERN,
            pd.DataFrame(
                {
                    "academic_year": ["2020-21", "2020-21", "2021-22"],
                    "academic_term": ["FALL", "SPRING", "FALL"],
                    "course_prefix": ["MATH", "MATH", "PHYS"],
                    "course_number": ["101", "202", "303"],
                    "course_cip": ["40.02", "45.06", "03.01"],
                    "course_type": ["CU", "CU", "CC"],
                    "delivery_method": ["O", "H", "F"],
                    "grade": ["4", "1", "W"],
                    "course_id": ["MATH101", "MATH202", "PHYS303"],
                    "course_subject_area": ["40", "45", "03"],
                    "course_passed": [True, False, pd.NA],
                    "course_completed": [True, True, False],
                    "course_level": [1, 2, 3],
                    "course_grade_numeric": [4.0, 1.0, pd.NA],
                    "course_grade": ["A", "D", "W"],
                }
            ).astype({"course_passed": "boolean", "course_grade_numeric": "Float32"}),
        ),
    ],
)
def test_add_course_features(df, min_passing_grade, course_level_pattern, exp):
    obs = course.add_features(
        df,
        min_passing_grade=min_passing_grade,
        course_level_pattern=course_level_pattern,
    )
    assert isinstance(obs, pd.DataFrame) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["df", "prefix_col", "number_col", "exp"],
    [
        (
            pd.DataFrame(
                {
                    "course_prefix": ["MATH", "BIO", "PHYS", "X", "Y", "Z"],
                    "course_number": ["101", "202", "303A", "404AB", "99", "001"],
                }
            ),
            "course_prefix",
            "course_number",
            pd.Series(["MATH101", "BIO202", "PHYS303A", "X404AB", "Y99", "Z001"]),
        ),
    ],
)
def test_course_id(df, prefix_col, number_col, exp):
    obs = course.course_id(df, prefix_col=prefix_col, number_col=number_col)
    assert isinstance(obs, pd.Series) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["df", "col", "min_passing_grade", "exp"],
    [
        (
            pd.DataFrame({"grade": ["4", "2", "P", "F", "0"]}, dtype="string"),
            "grade",
            1.0,
            pd.Series([True, True, True, False, False], dtype="boolean"),
        ),
        (
            pd.DataFrame(
                {"grade": ["4.0", "1.9", "P", "F", "2.1", "I", "W"]}, dtype="string"
            ),
            "grade",
            2.0,
            pd.Series([True, False, True, False, True, pd.NA, pd.NA], dtype="boolean"),
        ),
    ],
)
def test_course_passed(df, col, min_passing_grade, exp):
    obs = course.course_passed(df, col=col, min_passing_grade=min_passing_grade)
    assert isinstance(obs, pd.Series) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["df", "col", "exp"],
    [
        (
            pd.DataFrame(
                {"grade": ["4", "1", "P", "F", "0", "I", "W"]}, dtype="string"
            ),
            "grade",
            pd.Series([True, True, True, True, True, False, False]),
        ),
    ],
)
def test_course_completed(df, col, exp):
    obs = course.course_completed(df, col=col)
    assert isinstance(obs, pd.Series) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["df", "col", "pattern", "exp"],
    [
        (
            pd.DataFrame(
                {"course_number": ["101", "202", "303A", "404AB", "404ABC", "99", "1"]}
            ),
            "course_number",
            r"^(?P<course_level>\d)\d{2}(?:[A-Z]{,2})?$",
            pd.Series([1, 2, 3, 4, pd.NA, pd.NA, pd.NA], dtype="Int32"),
        ),
    ],
)
def test_course_level(df, col, pattern, exp):
    obs = course.course_level(df, col=col, pattern=pattern)
    assert isinstance(obs, pd.Series) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["df", "col", "exp"],
    [
        (
            pd.DataFrame({"grade": ["4", "2", "P", "F", "1"]}, dtype="string"),
            "grade",
            pd.Series([4.0, 2.0, pd.NA, pd.NA, 1.0], dtype="Float32"),
        ),
        (
            pd.DataFrame({"grade": ["4.0", "2.5", "I", "W", "1.75"]}, dtype="string"),
            "grade",
            pd.Series([4.0, 2.5, pd.NA, pd.NA, 1.75], dtype="Float32"),
        ),
    ],
)
def test_course_grade_numeric(df, col, exp):
    obs = course.course_grade_numeric(df, col=col)
    assert isinstance(obs, pd.Series) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty


@pytest.mark.parametrize(
    ["df", "grade_col", "grade_num_col", "exp"],
    [
        (
            pd.DataFrame(
                {
                    "grade": ["4", "2", "P", "F", "1"],
                    "grade_num": [4.0, 2.0, pd.NA, pd.NA, 1.0],
                },
            ).astype({"grade": "string", "grade_num": "Float32"}),
            "grade",
            "grade_num",
            pd.Series(["A", "C", "P", "F", "D"], dtype="string"),
        ),
        (
            pd.DataFrame(
                {
                    "grade": ["4.0", "3.3", "I", "W", "2.7", "0.7", "0.0", "A"],
                    "grade_num": [4.0, 3.3, pd.NA, pd.NA, 2.7, 0.7, 0.0, pd.NA],
                }
            ).astype({"grade": "string", "grade_num": "Float32"}),
            "grade",
            "grade_num",
            pd.Series(["A", "B", "I", "W", "B", "D", "F", "AUDIT"], dtype="string"),
        ),
    ],
)
def test_course_grade(df, grade_col, grade_num_col, exp):
    obs = course.course_grade(df, grade_col=grade_col, grade_num_col=grade_num_col)
    assert isinstance(obs, pd.Series) and not obs.empty
    assert obs.equals(exp) or obs.compare(exp).empty



import argparse
import logging
import pathlib
import sys

import faker
import pandas as pd

from student_success_tool.generation import pdp


def main():
    args = add_and_parse_arguments()

    logging.basicConfig(level=logging.INFO)

    faker.Faker.seed(args.seed)
    FAKER = faker.Faker()
    FAKER.add_provider(pdp.raw_cohort.Provider)
    FAKER.add_provider(pdp.raw_course.Provider)

    # institution_id must be the same for all records.
    institution_id = FAKER.numerify("#####!")
    student_guids = [FAKER.unique.student_guid() for _ in range(args.num_students)]
    cohort_records = [
        FAKER.raw_cohort_record(
            normalize_col_names=args.normalize_col_names,
            institution_id=institution_id,
            student_guid=student_guids[i],
        )
        for i in range(args.num_students)
    ]
    course_records = [
        FAKER.raw_course_record(
            cohort_record, normalize_col_names=args.normalize_col_names
        )
        for cohort_record in cohort_records
        for _ in range(
            FAKER.randomize_nb_elements(args.avg_num_courses_per_student, min=1)
        )
    ]
    df_cohort = pd.DataFrame(cohort_records)
    df_course = pd.DataFrame(course_records)

    logging.info(
        "generated %s cohort records and %s course records",
        len(cohort_records),
        len(course_records),
    )
    if args.save_dir is not None:
        args.save_dir.mkdir(parents=True, exist_ok=True)
        df_cohort.to_csv(
            args.save_dir / "INSTXYZ_STUDENT_SEMESTER_AR_DEIDENTIFIED.csv",
            header=True,
            index=False,
        )
        df_course.to_csv(
            args.save_dir / "INSTXYZ_COURSE_LEVEL_AR_DEID.csv",
            header=True,
            index=False,
        )
        logging.info("datasets saved to disk at %s", args.save_dir)


def add_and_parse_arguments():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument("--num-students", type=int, default=3)
    parser.add_argument("--avg-num-courses-per-student", type=int, default=5)
    parser.add_argument("--normalize-col-names", action="store_true", default=False)
    parser.add_argument("--save-dir", type=pathlib.Path, default=None)
    parser.add_argument("--seed", type=int, default=None)
    args = parser.parse_args()
    return args


if __name__ == "__main__":
    sys.exit(main())



import pytest
from unittest.mock import MagicMock
import pandas as pd
from student_success_tool.reporting.sections.registry import SectionRegistry
from student_success_tool.reporting.sections.data_sections import register_data_sections


@pytest.fixture
def mock_card_with_splits():
    card = MagicMock()
    card.cfg.preprocessing.splits = {"train": 0.7, "validate": 0.15, "test": 0.15}
    card.modeling_data = pd.DataFrame({"x": range(100)})
    card.format.bold.side_effect = lambda x: f"**{x}**"
    return card


@pytest.fixture
def mock_card_no_splits():
    card = MagicMock()
    card.cfg.preprocessing.splits = {}
    card.modeling_data = pd.DataFrame({"x": range(100)})
    card.format.bold.side_effect = lambda x: f"**{x}**"
    return card


def test_data_split_table_renders_correctly(mock_card_with_splits):
    registry = SectionRegistry()
    register_data_sections(mock_card_with_splits, registry)

    rendered = registry.render_all()
    output = rendered["data_split_table"]
    assert "| Split      | Students | Percentage |" in output
    assert "| Training   | 70       | 70%       |" in output
    assert "| Validation | 15       | 15%       |" in output
    assert "| Test       | 15       | 15%       |" in output


def test_data_split_table_missing_splits(mock_card_no_splits):
    registry = SectionRegistry()
    register_data_sections(mock_card_no_splits, registry)

    rendered = registry.render_all()
    assert rendered["data_split_table"] == "**Could not parse data split**"



import logging
import typing as t

LOGGER = logging.getLogger(__name__)


class SectionRegistry:
    """
    Collects and manages all registered model card sections.

    A SectionRegistry object allows section-generating functions to be registered under specific
    keys (e.g., "outcome_section") and later rendered into a markdown-friendly format. This
    design supports modularity, scale, and unit testing of individual sections of the model card.
    """

    def __init__(self):
        self._sections = []

    def register(
        self,
        key: str,
    ) -> t.Callable[[t.Callable[[], str]], t.Callable[[], str]]:
        """
        Registers a section-rendering function under a specific key.

        This is used as a decorator to attach a section generator (e.g. outcome section,
        bias table, etc.) to the registry. Each registered function is stored along with its
        key and later rendered via `render_all()`.

        Args:
            key: The identifier used to reference the section in the model card template.

        Returns:
            A decorator that registers the section-generating function.
        """

        def decorator(fn: t.Callable[[], str]) -> t.Callable[[], str]:
            self._sections.append((key, fn))
            return fn

        return decorator

    def render_all(self) -> dict[str, str]:
        """
        Renders all registered sections into a dictionary for markdown formatting.

        Returns:
            A mapping of section keys to their rendered markdown strings.
        """
        return {key: fn() for key, fn in self._sections}

    def clear(self) -> None:
        """
        Clears all registered sections for platform-specific processing & overrides.
        """
        self._sections.clear()
